/***************************************************************************
* blitz/alg/eigen.cc	  
* function	implementation of vectorized tridiagonal reduction and
* tridiagonal eigenvalue calculator
*
* $Id: eigen.cc,v 1.1.1.1 2003/07/09 23:55:52 mdsuresh Exp $
*
* Copyright (C) 2002 Suresh kumar Devanathan <mdsuresh@media.mit.edu>
* 
* $Id eigen.h Version 1.1
* Copyright (C) 1998-1999 Brent Worden	WordenWare <Brent@Worden.org>
*
* This	program	is free	software; you can redistribute it and/or
* modify it under the terms of	the	GNU	General	Public License
* as published	by the Free	Software Foundation; either	version	2
* of the License, or (at your option) any later version.
*
* This	program	is distributed in the hope that	it will	be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A	PARTICULAR PURPOSE.	 See the
* GNU General Public License for more details.
*
* Suggestions:			 blitz-dev@oonumerics.org
* Bugs:				 blitz-bugs@oonumerics.org
*
* For more	information, please	see	the	Blitz++	Home Page:
*	  http://oonumerics.org/blitz/
*
***************************************************************************
* $Log: eigen.cc,v $
* Revision 1.1.1.1  2003/07/09 23:55:52  mdsuresh
*
*
* Revision	1.0	 2002/10/26	20:22:49  mdsuresh
* Updated copyright date in headers.
**/


#ifndef	_VT_FUNCTIONS_EIGEN_CC
#define	_VT_FUNCTIONS_EIGEN_CC

#ifndef	_VT_PACK_H
#error the implementation of Eigenvalue solver requires inclusion of <blitz/alg/port.h> \
	prior to	the	inclusion of this file.
#endif	 

#include <blitz/alg/radius.cc>

namespace _VT_Pack {
	namespace Functions {
		namespace Matrix {
			namespace	EigenSystems {


				template<class T1, class T2>
					void  TridiagonalReduction(T1& a, T2& d, T2& e)
					//--------------------------------------------------------------------
					// Householder reduction of	a real,	symmetric matrix a.	 On	output,	a
					// is replaced by the orthogonal matrix	effecting the transformation.
					// d returns the diagonal elements of the tridiagonal matrix, and e
					// the off-diagonal	elements, with e[0]	= 0.
					//--------------------------------------------------------------------
				{
					int	n =	_VT_rows(a);
					_VT_PRECHECK( _VT_rows(a) == _VT_cols(a),
						"This operation is	only valid on square matrices.\n"
						"Invalid shape:" << _VT_shape(a) );
					//_VT_numtype(T1) zero(0);

					_VT_numtype(T1)	f, g, h,hh;

					_VT_Array1D( _VT_numtype(T1) ) gg( n);

					int	i, j, k, l;
                                        _VT_numtype(T1)	 scale,	tmp;
					_bz_use(i,j, k, l, tmp);
					_VT_resize(d,n);
					_VT_resize(e,n);
					_VT_DeclareRangeAll( _VT_RangeAll );

									
					for(i =	n -	1; i > 0; --i) {
						h = 0;
						l = i-1;

						if ( l > 0) {
						scale = _VT_sum(	_VT_abs(a(i,_VT_Range(0,l))) );
                        												
						if( scale ==	0 ) 
						   e(i) = a(i,l);
						else {
						
							_VT_Range k(0,l);
							a(i,k) /=	scale;
							h =	 _VT_sum(a(i,k) * a(i,k) ) ;
							
							f =	a(i,l);
							g =	-_VT_sign(sqrt(h),f);
							e(i) = scale * g;
							h -= f * g;
							a(i,l) = f - g;
							a(k,i) = a(i,k)/h;

													
							for(int	j =	0; j < i; ++j){
								    _VT_Range q(0,j);
								    _VT_Range p(j+1,l);
								    e(j) = ( _VT_sum( a(j,q) * a(i,q) )  +
										    _VT_sum( a(p,j) * a(i,p) ))/h;

								}

							f =	_VT_sum( e(k) *	a(i,k));
							hh =	f /	(h + h);
							e(k) -=	hh* a(i,k);
								
								for(int j =	0; j < i; ++j){
									
									_VT_Range k(0,j);
									a(j, k)	-=  a(i,j) * e(k)	+ e(j)* a(i,k);

								
								}
						}
						 
						}
						else  e(i)= a(i,l);
					     d(i) = h;
					}
					d(0) = 0;
					e(0) = 1;

					for(int i=0; i < n; i++)
					{
					  l = i-1;
					  _VT_Range q(0,l);
					  if(  d(i) != _VT_numtype(T1)(0.0) ) {
						  gg(q)  =  _VT_sum( a(i,q)(_VT_tensor::j)  * a(q,q)(_VT_tensor::j, _VT_tensor::i) , tensor::j);
						  a( q , q) -=  a(q,i)(_VT_tensor::i) * gg(q)( _VT_tensor::j);
					  }
					  d(i) = a(i,i);
					  a(i,i) = 1.0;
					  a(i,q) = 0;
					  a(q,i) = 0;

  

				 }
				}
						
			
			


				template<class T_, class T>
					void TridiagonalEigenvalues(T_& a, T& d, T& e)
					//--------------------------------------------------------------------
					// Algorithm to	determine the eigenvalues of a real, symetric,
					// tridiagonal matrix, or of a real, symmetric matrix previously
					// reduced by tridred().  On input,	d contains the diagonal	elements
					// of the tridaigonal matrix.  On output, it returns the eigenvalues.
					// e inputs	the	subdiagonal	elements of	the	tridiagonal	matrix,	with
					// e[0]	arbitrary.	On output, e is	destroyed.
					//--------------------------------------------------------------------
				{
					int n = _VT_rows(d);
					//_VT_numtype(T) zero(0), one(1), two(2);
					_VT_numtype(T) b, c,	f, g;

					_VT_PRECHECK( (_VT_rows(d) ==	_VT_rows(e)) &&
						(_VT_rows(a) ==   _VT_cols(a)) &&
						(_VT_rows(a) ==   _VT_rows(d)),
						"Incompatible shape mismatch\n"
						"d:" << _VT_shape(d) << "\n"
						"e:" << _VT_shape(e) << "\n" 
						"a:" << _VT_shape(a) << "\n" );

					//_VT_PRECHECK(	_VT_rows(d)	== _VT_rows(e),
					//					"This operation	is only	valid on square	matrices.\n"
					//				   "Invalid	shape:"	<< _VT_shape(a)	); 

					int i,	j, l, m;
					_VT_numtype(T) p, r,	s;
					_VT_numtype(T) tmp;

					_VT_DeclareRangeAll( _VT_RangeAll );
					_VT_Array1D( _VT_numtype(T) ) ff( n), _tmp(n);

					if(n != 1) {
						_VT_Range _i(1,	n-1);
						//cout << "e=" << e << "\n";
						_tmp = e;
						e(_i-1) = _tmp(_i);
						e(n - 1) =	0;
						//cout << "e=" << e << "\n";


						for(l = 0;	l <	n; ++l){
							j = 0;

							for(;;) {

								//_VT_Range _m(k, n-1);

								for( m= l; m < n-1; m++) {
									   tmp = _VT_abs(d(m)) + _VT_abs(d(m+1));
                                  if(  tmp == (_VT_abs( e(m) ) + tmp ) )
									   break;
								}

                              							
                                

								if( m == l) break;

								g = (d(l+1) - d(l)) / (2.0 * e(l));
								r = _VT_Pack::Helper::Radius(g, 1.0);
								g = d(m) - d(l) + e(l) / ( g + _VT_sign(r,g));
								s = 1.0;
								c = 1.0;
								p = 0.0;

								j++; 
								_VT_CONVERGE_FAIL(j <	50,	"Routine failed	to converage");	
								//throw	Exception("tridql",	"Routine failed	to converge.");


								for(i =	m -	1; i >=	l; --i){
									f =	s *	e(i);
									b =	c *	e(i);
									r =	_VT_Pack::Helper::Radius(f, g);
									e(i	+ 1) = r;
									if(r ==	0){
										d(i	+ 1) -=	p;
										e(m	) = 0;
										break;
									}
									s =	f /	r;
									c =	g /	r;
									g =	d(i	+ 1) - p;
									r =	(d(i) -	g) * s + c * 2 * b;
									p =	s *	r;
									d(i	+ 1) = g + p;
									g =	c *	r -	b;

									ff( _VT_RangeAll)  = a( _VT_RangeAll , i+1);
									a( _VT_RangeAll , i+1) = s* a( _VT_RangeAll, i) +
										c* ff(_VT_RangeAll);
									a( _VT_RangeAll, i) = c* a( _VT_RangeAll,i) -
										s* ff(_VT_RangeAll);


								}

								if ( r ==0.0 && i>=l) continue;


								d(l) -=	p;
								e(l) = g;
								e(m) = 0;

							}

						}

					
					}
				}






			}

		}
	}
}



#ifdef _VT_MATLAB_INTERFACE

namespace _VT_MATLAB_NAMESPACE {

	/* There are no	MATLAB equivalent names	for	these functions. */
	/* So they have	been adopted from the Numerical	Recipies book. */
/*
	template<class	T1,	class T2>
		_VT_rhs_<	_VT_Array2D( _VT_numtype(T1) ) , _VT_Array1D( _VT_numtype(T2) )	>
		tqli( T1	& _arr,	T2&	_e , T2 & _d)	 
	{
		_VT_Array1D(	_VT_numtype(T1)	) a	= _VT_copy(_arr);
		_VT_Array1D(	_VT_numtype(T2)	) e	= _VT_copy(_e);

		_VT_Pack::Functions::Matrix::EigenSystems::TridiagonalEigenvalues(a,e);
		return _VT_rhs( a, p);			

	}
	*/

	template<class T>
		_VT_rhs_<	_VT_Array2D( _VT_numtype(T)	) ,	_VT_Array1D( _VT_numtype(T)	)  >
		tridiag( T	& _arr)	 
	{
		_VT_Array2D(	_VT_numtype(T) ) arr = _VT_copy(_arr) ;
		_VT_Array1D(	_VT_numtype(T) ) a,p;
		_VT_Pack::Functions::Matrix::EigenSystems::TridiagonalReduction(arr,a,p);
		_VT_Pack::Functions::Matrix::EigenSystems::TridiagonalEigenvalues(arr,a,p);
        
		//cout << arr << " " << "a=" << a  << "p=" << p;

		return _VT_rhs( arr,  a);			 

	}

}

#endif // _VT_MATLAB_INTERFACE



#endif //  _BZ_NAMESPACE_EIGEN_CC
