/***************************************************************************
* blitz/alg/gaussj.cc     
* function implementation of vectorized gauss-jacobian matrix inversion
*
* $Id: gaussj.cc,v 1.1.1.1 2003/07/09 23:55:53 mdsuresh Exp $
*
* Copyright (C) 2002 Suresh kumar Devanathan <mdsuresh@media.mit.edu>
* 
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* Suggestions:          blitz-dev@oonumerics.org
* Bugs:                 blitz-bugs@oonumerics.org
*
* For more information, please see the Blitz++ Home Page:
*    http://oonumerics.org/blitz/
*
***************************************************************************
* $Log: gaussj.cc,v $
* Revision 1.1.1.1  2003/07/09 23:55:53  mdsuresh
*
*
* Revision 1.0  2002/10/26 20:22:49  mdsuresh
* Updated copyright date in headers.
**/

#ifndef _VT_FUNCTIONS_GAUSSJ_CC
#define _VT_FUNCTIONS_GAUSSJ_CC

#ifndef _VT_PACK_H
#error the implementation of gauss-jacobi inversion requires inclusion of <blitz/alg/port.h> \
	prior to the inclusion of this file.
#endif   

#include<algorithm>

namespace _VT_Pack {
	namespace  Functions {
		namespace Matrix {

			template<class T>
				void GaussJacobiInverse(T& a)
			{
				int n = _VT_rows(a);
				_VT_PRECHECK( _VT_rows(a) == _VT_cols(a),
					"This operation is only valid on square matrices.\n"
					"Invalid shape:" << _VT_shape(a) );

				//Range all;


				_VT_Array1D(int) indxc(n),indxr(n),ipiv(n);
				int i,icol,irow,j,k,l,ll;
				_VT_numtype(T) big,dum,pivinv,temp;
				_bz_use(i,j,  k ,temp);
				//for(j=0; j<n; j++) ipiv(j)= 0;
				indxr = 0;
				indxc = 0;
				ipiv=  0;
				_VT_Array2D( _VT_numtype(T) ) ipiv_selector(n,n);
				_VT_DeclareRangeAll( _VT_RangeAll );


				for(i=0; i<n;  i++) {
					big = 0.0;

					ipiv_selector   =  (1-ipiv(_VT_tensor::i))*(1- ipiv(_VT_tensor::j))   ;
					ipiv_selector *=   _VT_abs(a);

					_VT_PRECONDITION(  ! _VT_any( ipiv > 1) );
					_VT_Index2D index  =   
						_VT_maxIndex(ipiv_selector);
					big = _VT_abs(a(index));
					irow = index(0);
					icol = index(1);


					++(ipiv(icol));
					if(irow !=icol)
						_VT_swap(a(irow,_VT_RangeAll),a(icol,_VT_RangeAll));

					indxr(i) = irow;
					indxc(i) = icol;
					_VT_ILLCONDITION(a(icol,icol) != 0.0,
						"Cannot invert this ill-conditioned matrix.");

					pivinv = 1.0/a(icol,icol);
					a(icol,icol) =1.0;
					a(icol, _VT_RangeAll ) *=pivinv;

					// Todo: Needs work, parallization
					for( ll=0; ll<n; ll++){
						if(ll != icol){
							dum = a(ll,icol);
							a(ll,icol) = 0.0;
							a(ll,_VT_RangeAll) -= a(icol,_VT_RangeAll)*dum;
						}
					}
				}

				for( l=n-1; l >= 0; l--){
					//if(indxr(l) != indxc(l))
					//      for(k=0; k<n; k++)
					_VT_Range k(0,n-1);
					_VT_swap(a(k,indxr(l)),a(k,indxc(l)));
				}

			}

		}

	}
}
#ifdef _VT_MATLAB_INTERFACE

namespace _VT_MATLAB_NAMESPACE {
    using namespace _VT_Pack;

	template<class T>
		T gaussj(T _arr)
	{
		T arr = _VT_copy(_arr);
		_VT_Pack::Functions::Matrix::GaussJacobiInverse(arr);
		return arr;

	}
#ifdef  _VT_INVERSION_GUASS_JACOBI
	template<class T>
		T inv(T _arr)
	{
		return gaussj(_arr);
	}
#endif

}

#endif // _VT_MATLAB_INTERFACE

#endif // _VT_FUNCTIONS_GAUSSJ_CC
