/***************************************************************************
 * blitz/alg/mprove.cc     
 * function implementation of vectorized linear solution improver
 *
 * $Id: mprove.cc,v 1.1.1.1 2003/07/09 23:55:53 mdsuresh Exp $
 *
 * Copyright (C) 2002 Suresh kumar Devanathan <mdsuresh@media.mit.edu>
 * 
 * $Id eigen.h Version 1.1
 * Copyright (C) 1998-1999 Brent Worden WordenWare <Brent@Worden.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Suggestions:          blitz-dev@oonumerics.org
 * Bugs:                 blitz-bugs@oonumerics.org
 *
 * For more information, please see the Blitz++ Home Page:
 *    http://oonumerics.org/blitz/
 *
 ***************************************************************************
 * $Log: mprove.cc,v $
 * Revision 1.1.1.1  2003/07/09 23:55:53  mdsuresh
 *
 *
 * Revision 1.0  2002/10/26 20:22:49  mdsuresh
 * Updated copyright date in headers.
 **/
 
 
#ifndef _VT_FUNCTIONS_MPROVE_CC
#define _VT_FUNCTIONS_MPROVE_CC

#ifndef _VT_PACK_H
#error the implementation of mprove.cc requires inclusion of <blitz/alg/port.h> \
       prior to the inclusion of this file.
#endif   

#include <functions/lu.cc>


namespace _VT_Pack {
	namespace Functions {
		namespace	Matrix {
			namespace LinearSystems	{
				template<class T1, class T2, class T3>
					void ImprovedSolve(T1 a, T1 alud, T3 indx,	T2 b, T2& x)
				{

					int	j,i;
					//_bz_typename T1::T_numtype sdp;
					int	n =	_VT_rows(a);
					_VT_PRECHECK( _VT_rows(a) == _VT_cols(a),
						"This operation is only valid	on square matrices.\n"
						"Invalid shape:" << _VT_shape(a) );
					T2 r(n);

					r =	-b + _VT_sum(  a(_VT_tensor::i,	_VT_tensor::j) * x(_VT_tensor::j), _VT_tensor::j);

					/*	for	(i=1;i<=n;i++) {
					sdp	= -b[i];
					for	(j=1;j<=n;j++) sdp += a[i][j]*x[j];
					r[i]=sdp;
					}
					*/

					_VT_Pack::Functions::Matrix::Factorization::LUBackstitution(alud,indx,r);
					x -= r;
					//for (i=1;i<=n;i++) x[i] -= r[i];

				}
			}
		}
	}
}

#endif  // _VT_FUNCTIONS_MPROVE_CC
