/***************************************************************************
 * blitz/alg/svd.cc     
 * helper functions for calculation of SVDs, and Eigen values 
 * 
 *
 * $Id: svd.cc,v 1.1.1.1 2003/07/09 23:55:54 mdsuresh Exp $
 *
 * Copyright (C) 2002 Suresh kumar Devanathan <mdsuresh@media.mit.edu>
 * 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Suggestions:          blitz-dev@oonumerics.org
 * Bugs:                 blitz-bugs@oonumerics.org
 *
 * For more information, please see the Blitz++ Home Page:
 *    http://oonumerics.org/blitz/
 *
 ***************************************************************************
 * $Log: svd.cc,v $
 * Revision 1.1.1.1  2003/07/09 23:55:54  mdsuresh
 *
 *
 * Revision 1.0  2002/10/26 20:22:49  mdsuresh
 * Updated copyright date in headers.
 **/
 
#ifndef _VT_FUNCTION_SVD_CC
#define _VT_FUNCTION_SVD_CC

#ifndef _VT_PACK_H
#error the implementation of SVD requires inclusion of <blitz/alg/port.h> \
       prior to the inclusion of this file.
#endif

#include <blitz/alg/radius.cc>


namespace _VT_Pack {
	namespace  Functions {
		namespace Matrix {

			template<class T1, class T2>
				void   SVD(T1& a,  T2& w, T1& v)
			{
				/* BUG `nm' may be used uninitialized in this function */
				int m = _VT_rows(a), n = _VT_cols(a);
				bool  flag;
				int i, its,  jj, k, l, nm, nm1 = n - 1, mm1 = m - 1;
				_VT_numtype(T1)  c, f, h, s, x, y, z ,g = 0;
				_VT_numtype(T1)  anorm = 0.0,scale = 0.0;
				T2 tempn(n), tempm(m);;
			        _bz_use(flag, i, jj, l ,nm1, mm1);

				_VT_Array1D( _VT_numtype( T1) ) rv1(n);
				_VT_DeclareRangeAll( _VT_RangeAll );
				rv1 = 0;
				_VT_resize(w, _VT_cols(a));
				_VT_resize(v, _VT_cols(a), _VT_cols(a));
				w = 0;
				v = 0;
				_VT_PRECHECK(m <= n, "You must augment A with extra zero rows");

				/* Householder reduction to bidigonal form */
				for (i = 0; i < n; i++)
				{
					l = i + 1;
					rv1(i) = scale * g;
					g = s = scale = 0.0;
					if (i < m)
					{
						_VT_Range k(i,m-1);
						_VT_Range q(l,n-1);
						scale = _VT_sum(_VT_abs(a(k,i)));
						if (scale)
						{

							a(k,i) /= scale;
							s = _VT_sum(a(k,i) * a(k,i));

							f = a(i,i);
							g = -_VT_sign(_VT_sqrt(s), f);
							h = f * g - s;
							a(i,i) = f - g;
							
							tempn(q) =  _VT_sum( a(k, i)(_VT_tensor::j) 
								* a(k,q)(_VT_tensor::j, _VT_tensor::i), _VT_tensor::j) / h;
							a(k,q) = a(k,q) + a(k, i)(_VT_tensor::i) 
								* tempn(q)(_VT_tensor::j);
							a(k,i) *= scale;

						}
					}
						w(i) = scale * g;
						g = s = scale = 0.0;
						if (i < m && i != (n-1))
						{
							_VT_Range p(l,n-1);
							_VT_Range q(l,m-1);
							scale = _VT_sum(_VT_abs(a(i,p))); 

							if (scale)
							{
								a(i,p)/= scale;
								s = _VT_sum(a(i,p) * a(i,p));
								f = a(i,l);
								g = -_VT_sign(_VT_sqrt(s), f);
								h = f * g - s;
								a(i,l) = f - g;
								rv1(p) = a(i,p)/ h;
							
								tempm(q) = _VT_sum( a(q,p)(_VT_tensor::i, _VT_tensor::j) 
									* a(i,p)(_VT_tensor::j) ,_VT_tensor::j);
								a(q,p) +=  tempm(q)(_VT_tensor::i) * rv1(p)(_VT_tensor::j);
								a(i,p) *=scale;
							}
						}

						anorm = _VT_max(_VT_abs(w) + _VT_abs(rv1));
				}


				/* Accumulation of right-hand transformations */
				for (i = n - 1; i >= 0; i--)
				{
					if (i < n-1)
					{
						_VT_Range j(l,n-1);
						if (g)
						{
							/* double division to avoid possible underflow */
							
							v(j,i) = (a(i, j)  / a(i,l)) / g;
							tempn( j ) = _VT_sum(  a(i,j) (_VT_tensor::j) * v(j,j)(
								_VT_tensor::j,_VT_tensor::i) , _VT_tensor::j);
							v(j,j) += v(j,i)(_VT_tensor::i)* tempn(j)(_VT_tensor::j);
						}
						v(i,j) =0;
						v(j,i)= 0;

					}
					v(i,i) = 1.0;
					g = rv1(i);
					l = i;
				}
							/* Accumulation of left-hand transformations */
				for (i = _VT_min(n,m)-1; i >= 0; i--)
				{
					l = i + 1;
					g = w(i);
					_VT_Range j(l,n-1);
					a(i, j)  = 0;

					if (g)
					{
						g = 1.0 / g;
						_VT_Range q(i,m-1), p(l,n-1);
					
						tempn( p ) = (_VT_sum( a( _VT_Range(l,m-1) , i)(_VT_tensor::j) 
							* a( _VT_Range(l,m-1), p)
							( _VT_tensor::j, _VT_tensor::i), _VT_tensor::j)

							/ a(i,i)) * g;
						
						a( q , p ) += a(q,i)(_VT_tensor::i) * tempn(p)(_VT_tensor::j);
						a( q, i) *=g;


					} else
						a( _VT_Range(i,m-1),i) = 0;

					a(i,i) = a(i,i) + 1;
				}

				//cout << "rv1=" << rv1;

				/* diagonalization of the bidigonal form */
				for (k = n - 1; k >= 0; k--)
				{				/* loop over singlar values */
					for (its = 0; its < 30; its++)
					{	
						/* loop over allowed iterations */
						
						for (l = k; l >= 0; l--)
						{			/* test for splitting */
							nm = l - 1;	/* note that rv1[l] is always zero */
							if ((_VT_abs(rv1(l)) + anorm) == anorm)
								break;

							

							/*if (nm == -1) {
								//cout << _VT_abs(rv1(l));
								break;
							}*/
							

							if (((_VT_abs(w(nm)) + anorm) == anorm))
							{	

								c = 0.0;	/* cancellation of rv1[l], if l>1 */
								s = 1.0;
								for (i = 0; i <= k; i++)
								{
									f = s * rv1(i);
									rv1(i) = c* rv1(i);
									if ((_VT_abs(f) + anorm) == anorm) break;

									g = w(i);
									h = _VT_Pack::Helper::Radius(f, g);
									w(i) = h;
									h = 1.0 / h;
									c = g * h;
									s = (-f * h);
									_VT_Range j(0,m-1);

									tempm(j) = _VT_copy(a(j,nm));
									a(j,nm) = a(j,nm) * c+  a(j, i) * s;
									a(j,i) = -tempm(j)*s + a(j,i)*c;


								}
								break;
							}
						}
						z = w(k);
						if (l == k)
						{			/* convergence */
							if (z < 0.0)
							{
								w(k) = -z;
								_VT_Range j(0,n-1);
								v(j,k) = (-v(j,k));
							}
							break;
						}
						
						/*if( rv1(0) != 0  ) {
							   cout << rv1(0) << "\n";
							   assert(false);
							}
						*/
					
						_VT_PRECHECK(its < 30,
							"No convergence in SVD after 30 iterations");
						x = w(l);		/* shift from bottom 2-by-2 minor */
						nm = k - 1;
						y = w(nm);
						g = rv1(nm);
						h = rv1(k);
						f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2.0 * h * y);
						g = _VT_Pack::Helper::Radius(f, 1.0);
						/* next QR transformation */
						f = ((x - z) * (x + z) + h * ((y / (f + _VT_sign(g, f))) - h)) / x;
						c = s = 1.0;
						for (int j = l; j <= nm; j++)
						{
							i = j + 1;
							g = rv1(i);
							y = w(i);
							h = s * g;
							g = c * g;
							z = _VT_Pack::Helper::Radius(f, h);
							rv1(j) = z;

							/*if( rv1(0) != 0  && (j == 0) ) {
							   cout << "rv1(j)=" << rv1(j) << "\n";
							   //assert(false);
							}
							*/
						
							c = f / z;
							s = h / z;
							f = x * c + g * s;
							g = g * c - x * s;
							h = y * s;
							y = y * c;
							_VT_Range jj(0,n-1);
							tempn(jj) = _VT_copy(v(jj,j));
							v(jj,j) = v(jj,j)*c + v(jj,i)*s;
							v(jj,i) = -tempn(jj) *s + v(jj,i)* c;
							z = _VT_Pack::Helper::Radius(f, h);
							w(j) = z;	/* rotation can be arbitrary id z=0 */
							if (z)
							{
								z = 1.0 / z;
								c = f * z;
								s = h * z;
							}
							f = (c * g) + (s * y);
							x = (c * y) - (s * g);
							jj = _VT_Range(0,m-1);
							tempm(jj)= a(jj,j);
							a(jj,j) = a(jj,j) * c + a(jj,i)*s;
							a(jj,i) = -tempm(jj)*s +a(jj,i)*c;


						}
						rv1(l) = 0.0;
						rv1(k) = f;
						/*if( rv1(0) != 0  ) {
							   cout << "f="  << f << "\n";
							   cout << "k=" << k << "\n";

							   cout << rv1(0) << "\n";
							   assert(false);
							}
							*/
						

						//rv1(0) = 0;  // This line is an unwanted BUG fix

						//if( rv1(0) != 0 )  assert(false);
						w(k) = x;
					}
				}

							//    cout << "A=" << a << "\n"
							//         << "B=" << w << "\n"
							//        << "W=" << v ;

			}
				

			template<class T1, class T2, class T3> 
				void SVDBackstitution(T1 u, T2 w, T1 v, T3 b, T3& x)
			{
				int jj,j,i;

				int m = _VT_rows(u);
				int n = _VT_cols(u);
				_VT_numtype(T1) s;
				T2 temp(n);

				temp =   _VT_sum(u(_VT_tensor::i, _VT_tensor::j)
					*b(_VT_tensor::i), _VT_tensor::i);
				temp =   _VT_where( w!=0 , temp * _VT_pow(w,-1), 0); 


				/*for (j=1;j<=n;j++) {
				s=0.0;
				if (w(j)) {


				for (i=1;i<=m;i++) s += u[i][j]*b[i];
				s /= w[j];
				}
				tmp[j]=s;
				}
				*/

				x = _VT_sum( v( _VT_tensor::j, _VT_tensor::i) 
					* temp( _VT_tensor::i), _VT_tensor::i);

				/*for (j=1;j<=n;j++) {
				s=0.0;
				for (jj=1;jj<=n;jj++) s += v[j][jj]*tmp[jj];
				x[j]=s;
				}
				free_vector(tmp,1,n);
				*/
			}

			template<class T1, class T2, class T3>
				void   SVD(T1& a,T2& w, T1& v, T3 condition)
			{
				int     j;
				_VT_numtype(T1)  wmax, thresh;
				_bz_use(j);
                                SVD(a,w,v);
				/* Editing the sigular values given condition */
				wmax = _VT_max(w);
				thresh = condition * wmax;
				w = _VT_where( (w) < thresh, 0 , w);

		 }



			template<class T1> 
				T1 PesudoInverse(T1 u, _VT_numtype(T1) condition)
			{
				T1 s = _VT_copy(u),d;
				_VT_Array1D( _VT_numtype(T1) ) v, inv_v;

				SVD(s,v,d, condition);
				//inv_v= Inverse(v);

				//cout << v << "\n";

				v = _VT_where( _VT_abs(v)!=0,    _VT_pow(v,-1), 0);
				// cout << v << "\n";

				T1 v_mat( _VT_rows(v), _VT_rows(v)); 
				v_mat =0;
				for( int i=0; i < _VT_rows(v); i++)
					v_mat(i,i) = v(i);

				//  cout <<  s << "\n" << v_mat << "\n" << d << "\n";
				// cout << "tranpose";
				T1 r( _VT_shape(s) ), q( _VT_shape(s) );
				r= 0;
				q = 0;

				//cout << d;

				r =(_VT_sum(d(_VT_tensor::i, _VT_tensor::k) * v_mat(_VT_tensor::k,_VT_tensor::j),_VT_tensor::k));
				q = _VT_sum(r(_VT_tensor::i, _VT_tensor::k) * s    (_VT_tensor::j,_VT_tensor::k),_VT_tensor::k);
				return q;



			}


		}
	}

}
#ifdef _VT_MATLAB_INTERFACE

namespace _VT_MATLAB_NAMESPACE {
   using namespace _VT_Pack;

    template<class T>
     T pinv( T _input, _VT_numtype(T) condition)
     {
        return  _VT_Pack::Functions::Matrix::PesudoInverse( _input, condition);
     }
     
    
     
   template<class T>
   _VT_rhs_< _VT_Array2D( _VT_numtype(T) ) , 
             _VT_Array1D( _VT_numtype(T) ), 
             _VT_Array2D( _VT_numtype(T) ) >  svd(T _input)
     {
         _VT_Array2D( _VT_numtype(T) ) s =  _VT_copy(_input);
         _VT_Array1D( _VT_numtype(T) ) v ;
         _VT_Array2D( _VT_numtype(T) ) d;
         _VT_Pack::Functions::Matrix::SVD(s,v,d);
         return _VT_rhs( s , v, d);
    }
    
   template<class T>
   _VT_rhs_< _VT_Array2D( _VT_numtype(T) ) , 
             _VT_Array1D( _VT_numtype(T) ), 
             _VT_Array2D( _VT_numtype(T) ) > 
     svd(T _input, _VT_numtype(T) condition)
    {
      _VT_Array2D( _VT_numtype(T) ) s =  _VT_copy(_input);
      _VT_Array1D( _VT_numtype(T) ) v ;
      _VT_Array2D( _VT_numtype(T) ) d;
      _VT_Pack::Functions::Matrix::SVD(s,v,d, condition);
      return _VT_rhs( s , v , d);
   }
     
     
}
#endif // _VT_MATLAB_INTERFACE

#endif // _VT_FUNCTION_SVD_CC
