/***************************************************************************
 * blitz/alg/vander.cc     
 * implementation of vectorized vandermode interpolation
 * 
 * $Id: vander.cc,v 1.1.1.1 2003/07/09 23:55:54 mdsuresh Exp $
 *
 * Copyright (C) 2002 Suresh kumar Devanathan <mdsuresh@media.mit.edu>
 * 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Suggestions:          blitz-dev@oonumerics.org
 * Bugs:                 blitz-bugs@oonumerics.org
 *
 * For more information, please see the Blitz++ Home Page:
 *    http://oonumerics.org/blitz/
 *
 ***************************************************************************
 * $Log: vander.cc,v $
 * Revision 1.1.1.1  2003/07/09 23:55:54  mdsuresh
 *
 *
 * Revision 1.0  2002/10/26 20:22:49  mdsuresh
 * Updated copyright date in headers.
 **/
 
#ifndef _VT_FUNCTION_VANDER_CC
#define _VT_FUNCTION_VANDER_CC

#ifndef _VT_PACK_H
#error the implementation of Vandermode matrix solvers requires inclusion of <blitz/alg/port.h> \
	prior to the inclusion of this file.
#endif

namespace _VT_Pack
{
	namespace  Functions {

		namespace LinearSystems {
			template<class T>
				T Vandermonde(T x, T q)
			{
				int i,j,k,k1;
               			_VT_numtype(T) b,s,t,xx;
				_bz_use( i, j, k, b, s,t, xx, k1);
				_VT_PRECHECK( _VT_rows(x) == _VT_rows(q),
					"Invalid shape mismatch\n"
					"x:" << _VT_shape(x)  <<
					"q:" << _VT_shape(q)  );

				int n = _VT_rows(x);
				_VT_Array2D( _VT_numtype(T) ) a(n,n);

				T w(n),c(n);
				if (n == 1) return q;
				else {
					c = 0;
					c(n-1) = -x(0);
					for (i=1;i<n;i++) {
						_VT_Range _j(n-i-1,n-2);
						c(_j) = c(_j) - x(i)*c(_j+1);
						c(n-1) = c(n-1) - x(i);
					}

					//cout << "c=" << c;

					a = x(_VT_tensor::i) - x(_VT_tensor::j);
		            a = _VT_where( _VT_tensor::i == _VT_tensor::j , 1, a);

					//cout << a;

					w =  _VT_product(a, _VT_tensor::j);

					//cout << w;
					a =  _VT_where( _VT_tensor::i == _VT_tensor::j , 0, a);

					_VT_DeclareRangeAll( _VT_RangeAll );

					//cout << x;

					a( _VT_RangeAll, 0) = -c(0) * _VT_pow(x,-1);

					//cout <<  a(_VT_RangeAll,0);

					for(int i=1; i < n; i++)
					{
						a( _VT_RangeAll,i) = -(c(i) - a( _VT_RangeAll,i-1)) * _VT_pow(x,-1);
					}
					x =  _VT_sum( a( _VT_tensor::i, _VT_tensor::j) * q(_VT_tensor::j), _VT_tensor::j);

					w = x * _VT_pow(w,-1);
					return w;

				}
			}
		}
	} 
}

#ifdef _VT_MATLAB_INTERFACE

namespace _VT_MATLAB_NAMESPACE {
   using namespace _VT_Pack;

	/* vandermode polynomial coefficient fitting */
	template<class T>
	T vander(T x, T y) 
	{ 
		_VT_Array1D(_VT_numtype(T) ) x_ = _VT_copy(x);
		_VT_Array1D(_VT_numtype(T) ) y_ = _VT_copy(y);
		return _VT_Pack::Functions::LinearSystems::Vandermonde(x_,y_);
	}

}

#endif // _VT_MATLAB_INTERFACE

#endif // _VT_FUNCTION_VANDER_CC
