#include "all.h"

void update(int gate,   /* which gate in circuit to work on */
	    int in_out) /* 0=inputlist,1=outputlist */
{ /* update */
  int i;
  int temp;
  
  if(in_out == 0)  /* update input gate list */
    {
      for(i=1;i<=circuit[gate]->inputs[0];i++)
	{
	  if(circuit[gate]->inputs[i] > num_pi)
	    circuit[gate]->inputs[i] = ref_tab[circuit[gate]->inputs[i]];
	}
    }
  else if(in_out == 1)
    {
      for(i=1;i<=circuit[gate]->outputs[0];i++)
	{
	  if(scircuit[circuit[gate]->outputs[i]]->funct == dff)
	    {
	      temp = circuit[gate]->outputs[i];
	      circuit[gate]->outputs[i]= gatecnt -num_pi + (int) ref_tab[temp];
	    }
	  else if(circuit[gate]->outputs[i] > num_pi && 
		  circuit[gate]->outputs[i] < gatecnt-num_po &&
		  scircuit[circuit[gate]->outputs[i]]->funct != dff)
	    circuit[gate]->outputs[i] = 
	      ref_tab[circuit[gate]->outputs[i]];
	}
    }
  else
    {
      fprintf(resultsout,
	      "neither input nor output list was specified to work on\n");
      exit(-1);
    }
} /* update */

void printoutfunct(nodes *ckt, int agate)
{ /* printoutfunct */
  int gatefunct;
  
  gatefunct = ckt[agate]->funct;
  
  switch(gatefunct)
    {
    case and:
      fprintf(resultsout,"%d an ",agate);
      break;
    case nand:
      fprintf(resultsout,"%d na ",agate);
      break;
    case or:
      fprintf(resultsout,"%d or ",agate);
      break;
    case nor:
      fprintf(resultsout,"%d no ",agate);
      break;
    case not:
      fprintf(resultsout,"%d in ",agate);
      break;
    case xor:
      fprintf(resultsout,"%d xo ",agate);
      break;
    case equiv:
      fprintf(resultsout,"%d eq ",agate);
      break;
    case pi:
      fprintf(resultsout,"%d pi ",agate);
      break;
    case po:
      fprintf(resultsout,"%d po ",agate);
      break;
    case dff:
      fprintf(resultsout,"%d df ",agate);
      break;
    default:
      break;
    }
} /* printoutfunct */

void cktconvert (void)
{ /* cktconvert */
  int gate;
  int i;
  int index;
  int in = 0;
  int out = 1;

  if(seq_circuit == true)
    {
      circuit=(nodes *) malloc((gatecnt + 2 + num_dff) * sizeof(nodes));
      if(circuit == NULL)
	{
	  printf("no memory for circuit graph. Abort.\n");
	  fprintf(resultsout,"no memory for circuit graph. Abort.\n");
	  exit(-1);
	}
      totalcktmemory += (gatecnt+3+num_dff) * sizeof(nodes);
      init_node(circuit,gatecnt+num_dff+1);
  
/* building dfflist */
      dfflist = make_ilist(num_dff);
      if(dfflist == NULL)
	{
	  printf("no memory for dfflist. Abort.\n");
	  fprintf(resultsout,"no memory for dfflist. Abort.\n");
	  exit(-1);
	}
      totalcktmemory += (num_dff+1)*sizeof(int);
      dfflist[0]=0;
      for(gate=1;gate<=gatecnt;gate++)
	{
	  if(scircuit[gate]->funct == dff)
	    {
	      dfflist[0]++;
	      if(dfflist[0] > num_dff)
		{
		  printf("Inconsistent number of DFF's. Abort.\n");
		  fprintf(resultsout,
			  "Inconsistent number of DFF's. Abort.\n");
		  exit(-1);
		}
	      else
		dfflist[dfflist[0]]=gate;
	    }
	}
/*
      printf("Print dfflist: dff# -> gate#\n");
      for(i=1;i<=dfflist[0];i++)
	printf("%d -> %d.\n",i,dfflist[i]);
      exit(0);
*/      
/* print out the scircuit */
/*
#ifdef DEBUG
      fprintf(resultsout,"gate input list in scircuit \n");
      for(gate=1;gate<=gatecnt;gate++)
	{
	  printoutfunct(scircuit,gate);
	  loop_through_ilist(scircuit[gate]->inputs,index,ingate)
	    fprintf(resultsout,"%d ",ingate);
	  fprintf(resultsout,"\n");
	}      
      fprintf(resultsout,"gate output list in scircuit\n");
      for(gate=1;gate<=gatecnt;gate++)
	{
	  printoutfunct(scircuit,gate);
	  loop_through_ilist(scircuit[gate]->outputs,index,ingate)
	    fprintf(resultsout,"%d ",ingate);
	  fprintf(resultsout,"\n");
	}      
#endif
*/
      
/* open reference table */
      ref_tab = (int *)malloc((gatecnt+1)*sizeof(int));
      ref_tab[0] = gatecnt /* -num_po */;
/* build up reference table for gates: id in scircuit <=> id in curcuit */
      i = 1;
      index = num_dff;
      for(gate=num_pi+1;gate<=gatecnt-num_po;gate++)
	{
	  if(scircuit[gate]->funct == dff)
	    {
	      ref_tab[gate]=num_pi+i;
	      i++;
	      index--;
	    }
	  else
	    ref_tab[gate]=gate+index;
	}
      for(i=gatecnt-num_po+1;i<=gatecnt;i++)
	ref_tab[i] = i;
      for(i=1;i<=num_pi;i++)
	ref_tab[i] = i;
/*
#ifdef DEBUG
      printf("Print ref_tab:sckt-gate -> ckt-gate\n");
      for(i=1;i<=gatecnt;i++)
	{
	  printf("%d -> %d\n",i,ref_tab[i]);
	}
#endif      
*/
/* report error is number of DFFs not match */
      if(index != 0)
	{
	  fprintf(resultsout,"number of DFFs in dfflist not = in scircuit\n");
	  exit(-1);
	}
/* move in PIs from scircuit to circuit */
      for(gate=1;gate<=num_pi;gate++)
	{
	  if(scircuit[gate]->funct != pi)
	    {
	      fprintf(resultsout,"PI id error\n");
	      exit(-1);
	    }
	  else
	    {
	      nodecopy(gate, gate);
	      if(circuit[gate]->outputs != NULL)
		update(gate,out);
	      else
		{
		  fprintf(resultsout,"gate %d outputlist empty\n",gate);
		  exit(-1);
		}
	    }
	}
  
 /* build PPIs for circuit */
      for(i=1,gate=num_pi+1;gate<=num_pi+num_dff;gate++,i++)
	{
	  if(scircuit[dfflist[i]]->funct != dff)
	    {
	      fprintf(resultsout,"DFF id error -> dfflist error\n");
	      exit(-1);
	    }
	  else
	    {
	      nodecopy(gate, dfflist[i]);
	      circuit[gate]->funct = pi;
	      circuit[gate]->inputs = NULL;      
	    }
	  if(circuit[gate]->outputs != NULL)
	    update(gate,out);
	  else
	    {
	      fprintf(resultsout,"gate %d outputlist empty\n",dfflist[i]);
	      exit(-1);
	    }
	}
/* move in gates (except DFF) and POs from scircuit to circuit */
      for(i=1,gate=num_pi+num_dff+1;gate<=gatecnt;i++)
	{
	  if(scircuit[num_pi+i]->funct != dff)
	    {
	      nodecopy(gate, num_pi+i);
	      if(circuit[gate]->inputs != NULL)
		{
		  update(gate,in);
		  gate++;
		}
	      else
		{
		  fprintf(resultsout,"gate %d inputlist is empty\n",num_pi+i);
		  exit(-1);
		}
	      if(circuit[gate-1]->outputs != NULL)
		update(gate-1,out);
	      else
		;
	    }
	  else
	    ;
	}
/* build PPOs for circuit */
      for(i=1,gate=gatecnt+1;gate<=gatecnt+num_dff;gate++,i++)
	{
	  if(scircuit[dfflist[i]]->funct != dff)
	    {
	      fprintf(resultsout,"DFF id error -> dfflist\n");
	      exit(-1);
	    }
	  else
	    {
	      nodecopy(gate, dfflist[i]);
	      circuit[gate]->funct = po;
	      circuit[gate]->outputs = NULL;
	      if(circuit[gate]->inputs != NULL)
		update(gate,in);
	      else
		{
		  fprintf(resultsout,
			  "gate %d inputlist is empty\n", dfflist[i]);
		  exit(-1);
		}
	    }
	}
      
      for (index=1;index<=gatecnt+num_dff;index++)
	{
	  if (inputlist(index)!=NULL)
	    {
	      i = ilist_length(inputlist(index));
	      circuit[index]->lines[0] = (int *)make_ilist(i);
	      circuit[index]->lines[1] = (int *)make_ilist(i);
	      ilist_length(circuit[index]->lines[0]) = i;
	      ilist_length(circuit[index]->lines[1]) = i;
	    }
	}

/* print out the circuit */

#ifdef DEBUG
      fprintf(resultsout,"gate input list in circuit\n");
      for(gate=1;gate<=gatecnt+num_dff;gate++)
	{
	  printoutfunct(circuit,gate);
	  if(circuit[gate]->inputs != NULL)
	    for(i=1;i<=circuit[gate]->inputs[0];i++)
	      fprintf(resultsout,"%d ",circuit[gate]->inputs[i]);
	  fprintf(resultsout,"\n");
	}     
/*
      fprintf(resultsout,"gate output list in circuit\n");
      for(gate=1;gate<=gatecnt+num_dff;gate++)
	{
	  printoutfunct(circuit,gate);
	  if(circuit[gate]->outputs != NULL)
	    for(i=1;i<=circuit[gate]->outputs[0];i++)
	      fprintf(resultsout,"%d ",circuit[gate]->outputs[i]);
	  fprintf(resultsout,"\n");
	}      
*/
#endif

      gatecnt += num_dff;
      num_po += num_dff;
      num_pi += num_dff;

      cckt_to_sckt = make_ilist(gatecnt+1);
      cckt_to_sckt[0] = 0;
      cckt_to_sckt[gatecnt+1] = gatecnt+1;

      for(i=1;i<=num_pi-num_dff;i++)
	cckt_to_sckt[i] = i;
      for(i=num_pi-num_dff+1;i<=gatecnt-num_po;i++)
	cckt_to_sckt[ref_tab[i]] = i;
      for(i=gatecnt-num_po+1;i<=gatecnt-num_dff;i++)
	cckt_to_sckt[i] = i;
      for(i=gatecnt-num_dff+1;i<=gatecnt;i++)
	cckt_to_sckt[i] = dfflist[i-gatecnt+num_dff];
/*
      printf("Print cckt-sckt ref_tab: cckt-gate -> sckt-gate\n");
      for(i=1;i<=gatecnt;i++)
	printf("%d -> %d\n",i,cckt_to_sckt[i]);
      exit(0);
*/
    }
  else
    {
      circuit = scircuit;

#ifdef DEBUG
      fprintf(resultsout,"gate input list in circuit\n");
      for(gate=1;gate<=gatecnt+num_dff;gate++)
	{
	  printoutfunct(circuit,gate);
	  if(circuit[gate]->inputs != NULL)
	    for(i=1;i<=circuit[gate]->inputs[0];i++)
	      fprintf(resultsout,"%d ",circuit[gate]->inputs[i]);
	  fprintf(resultsout,"\n");
	}     
#endif

/* open reference table */
      ref_tab = (int *)malloc((gatecnt+1)*sizeof(int));
      ref_tab[0] = gatecnt;
/* build up reference table for gates: id in scircuit <=> id in curcuit */
      for(i=1;i<=gatecnt;i++)
	ref_tab[i] = i;

      for (index=1;index<=gatecnt+num_dff;index++)
	{
	  if (inputlist(index)!=NULL)
	    {
	      i = ilist_length(inputlist(index));
	      circuit[index]->lines[0] = (int *)make_ilist(i);
	      circuit[index]->lines[1] = (int *)make_ilist(i);
	      ilist_length(circuit[index]->lines[0]) = i;
	      ilist_length(circuit[index]->lines[1]) = i;
	    }
	}
    }
} /* cktconvert */
