/***********************************************************************
  implicat_forw.c

  Change Log
  
  Date     Ini Comment
  -------- --- -------
  02/13/92 PC  Change implicate_forw to take input from input lines,
               instead of input nodes
 ***********************************************************************/

#include "all.h"

/*
 * Function:                                                             
 *   Do forward implication according to the supplied inputs and the gate
 *   type.
 * Input:                                                                
 *   gateno - inputs to the gate                                         
 *   gatetp - the gate type                                              
 * Return:                                                               
 *   none                                                                
 * SideEffect:                                                           
 *   sets label0 and label1 of the gate                                
 * PeteChen 2/13/92
 */

void implicate_forw(NODE *gateno, int gatetp)
{ /* implicate_forw */
  int i;            /* index for looping thorough inputlist */
  int result0;      /* 0 word of the result */
  int result1;      /* 1 word of the result */
  int operand0;     /* 0 word of one of the binary operand */
  int operand1;     /* 1 word of one of the binary operand */
  int temp0;        /* 0 and 1 word of a temp */

  for(i=1; i<=nmbrinputs(gateno);i++)
    {
      if (i == 1)
	{
	  result0 =  get_line0(gateno,i-1);
	  result1 =  get_line1(gateno,i-1);
	}
      else
	{
	  operand0 = get_line0(gateno,i-1);
	  operand1 = get_line1(gateno,i-1);
	  switch(gatetp)
	    {
	    case p_and: case p_nand:
	      result0 |= operand0;
	      result1 &= operand1;
	      break;
	    case p_or: case p_nor:
	      result0 &= operand0;
	      result1 |= operand1;
	      break;
	    case p_xor: case p_equiv:
	      temp0   = (result0 | operand1) & (result1 | operand0);
	      result1 = (result1 & operand0) | (result0 & operand1);
	      result0 = temp0;
	      break;
	    }
	}
    }
  switch(gatetp)
    {
    case p_ppo:
      result0 = result0;
      result1 = result1;
      break;
    case p_nand: 
    case p_nor: 
    case p_not: 
      //temp0   = result0;
      //result0 = result1;
      //result1 = temp0;
      //break;
    case p_equiv:
      temp0   = result0;
      result0 = result1;
      result1 = temp0;
      break;
    }
  get_label0(gateno) = result0;
  get_label1(gateno) = result1;
} /* implicate_forw */


