/*
 * sfsim.c
 *   routines for actual sequential fault simulation
 *
 * PeteChen
 *
 *
 */

#include "all.h"

static NODE	**eventqueue;
static int  ytail = 0; /* Count of events in the event queue */
static int  yhead = 1; /* Used to loop through event queue */
static int  yempty = 1; /* Flag indicating whether queue is empty or full */
static int equeue_allocated = 0;
static int queuesize;
//#define QUEUESIZE 325 * num_pi
#define QUEUESIZE 325 * num_pi
//#define FCLISTSIZE 5*num_pi
#define FCLISTSIZE 5*num_pi
char reading [161]; /* Input buffer for fsim */

/*gatecnt = #pi+#d_ff+#non_ff_gate+#po+#d_ff.  (it includes PPI and PPO)*/

#define cfaultto(gate) \
  ((gate==0)? 0: \
   ((gatefunction(get_node_pntr(circuit, gate))==p_d_ff)? \
    gatecnt-num_dff+ref_tab[gate]-(nmbr_pi(circuit)): \
    ref_tab[gate]))

void print_fdetected(int updatefmode)   
{ 
  int i, j;

  j = 0;   
  fprintf(resultsout,"Printing faults detected(print_fdetected):----- \n ");

  for(i=1;i<=faultcnt;i++){                     
    if(updatefmode != UPDATE_FAULT)                       
      {                                           
	if(faultlist[i]->flag == true)             
	  { 
	    fprintf(resultsout, "Fault[%d] %5d %5d %2d\n", 
		    i, faultlist[i]->from, faultlist[i]->to, faultlist[i]->sa);
	    j++;
	  } 
      }                                           
    else                                          
      {                                           
	if(tmpdetected[i] == true)                
	  { 
	    fprintf(resultsout, "Fault[%d] %5d %5d %2d\n", 
		    i, faultlist[i]->from, faultlist[i]->to, faultlist[i]->sa);
	    j++;
	  } 
      }                                           
  }                                               
  fprintf(resultsout,"total # faults detected = %d \n",j);
}

#define event_enqueue(number) do \
  {                                                      \
    ytail = (ytail + 1)   % (queuesize);                 \
    if (yempty)                                          \
      yempty = 0;					 \
    else if (ytail == yhead)				 \
      {							 \
	fprintf(resultsout, "Event queue overflow.\n"); \
	exit (-1);					 \
      }							 \
    *(eventqueue+ytail) = (number);			 \
  } while (0)

#define event_dequeue(number) do \
  {                                           \
    (number) = *(eventqueue+yhead);	      \
    yhead = (yhead + 1) % (queuesize);	      \
    if (((ytail + 1) % (queuesize)) == yhead) \
      yempty = 1;			      \
  } while (0)

#define empty_queue() yempty
#define set_bit(w,b) ((w) |= 01<<(b))
#define clear_bit(w,b) ((w) &= ~(01<<(b)))

/*
 * Function:
 *   set a particular bit of a line (a word) to a value
 * Input:
 *   gate - the gate number 
 *   line - the line number to be set, 0 refers to the label instead
 *   bit  - the ordinal number of the bit to be set (0-31)
 *   val  - value to set to (0 or 1)
 * Return:
 *   none
 */
void set_line (NODE *gate, /* gate number */
	       int line, /* input line number, 0 is label */
	       int bit,  /* which bit to set */
	       int val)  /* value to set to */
{ /* set_line */
  if (line==0)
    {
      if (val==0)
	{
	  set_bit(get_label0(gate),bit);
	  clear_bit(get_label1(gate),bit);
	}
      else
	{
	  clear_bit(get_label0(gate),bit);
	  set_bit(get_label1(gate),bit);	    
	}
    }
  else
    {
      if (val==0)
	{
	  set_bit(get_line0(gate,line-1),bit);
	  clear_bit(get_line1(gate,line-1),bit);
	}
      else
	{
	  clear_bit(get_line0(gate,line-1),bit);
	  set_bit(get_line1(gate,line-1),bit);	    
	}
      
    }
} /* set_line */

/*print the states of the circuit
fault_ff:    0 - print in the order of faults
             1 - print in the order of FFs
ff_num:     -1 - print states of all FFs
            otherwise print states of the specified FF  
fault_num:  -1 - print states of all faults
            otherwise print states of the specified fault  
ft:         local fault list
inpt:        1 - states of FF inputs 
             0 - states of FF outputs 
*/
void print_state(int fault_ff, int ff_num, int fault_num, int *ft, int inpt)
{
  int m,vec;
  int ppo_start, po_index;
  
  if(inpt == true) ppo_start = nmbr_pi (circuit)+1;
  else ppo_start = gatecnt-num_dff+1;

  if(fault_ff==0) //print in the order of faults
    {
      fprintf(resultsout, "Printing local fault states(print_state)--------\n");
      for(vec=0;vec<32;vec++)
      {
	if(fault_num == -1 ||fault_num == vec)
	  {
	    if(vec==0)
	      fprintf(resultsout, "      good FF states---");
	    else 
	      fprintf(resultsout,"Fault[%2d-%3d] %5d %5d %2d states---", 
		vec, ft[vec],
		faultlist[ft[vec]]->from,faultlist[ft[vec]]->to,
		faultlist[ft[vec]]->sa);
	    //	      fprintf(resultsout, "faulty states[%2d]---",vec);
	    /*MB need to set nmbr_ppo
		    need to create ppo_list
		    need to create get_label0 and get_label1 */
	    for(m=0; m</*nmbrppos(circuit)*/num_dff; m++)
	      {
		po_index = ((get_label0(circuit->Ppo_list[m])>>vec)&01) |
		  (((get_label1(circuit->Ppo_list[m])>>vec)<<1)&02);
		fprintf(resultsout, "%s", trans_state[(int)po_index]);	  
	      }
	    fprintf(resultsout,"\n");
	  }//if(fault_num...
      }
    }
  
  if(fault_ff==1) //print in the order of faults
    {
      fprintf(resultsout, "Printing FF states(print_state)--------\n");
      for(m=0; m</*nmbrppos(circuit)*/num_dff; m++)
      {
	if(ff_num == -1 || ff_num == m)
	  {
	    fprintf(resultsout, "FF[%3d] states---",m-ppo_start);
	    for(vec=0;vec<32;vec++)
	      {
		po_index = ((get_label0(circuit->Ppo_list[m])>>vec)&01) |
		  (((get_label1(circuit->Ppo_list[m])>>vec)<<1)&02);
		fprintf(resultsout, "%s", trans_state[(int)po_index]);	  
	      }
	    fprintf(resultsout,"\n");
	  }//if(ff
      }
    }
}

void clearLabels(CIRCUIT *circuit)
{
    int i;
    for(i=0;i<nmbr_pi(circuit);i++)
	PIs(circuit, i)->label0 = PIs(circuit, i)->label1 = 0;

    for(i=0;i<nmbr_gates(circuit);i++)
        gatesx(circuit, i)->label0 = gatesx(circuit, i)->label1 = 0;

    for(i=0;i<nmbr_po(circuit);i++)
        POs(circuit, i)->label0 = POs(circuit, i)->label1 = 0;

    if(nmbrppos(circuit))
    {
        for(i=0;i<nmbrppos(circuit);i++)
            PPOs(circuit, i)->label0 = PPOs(circuit, i)->label1 = 0;
    }

} /*end of clearLabels()*/

/*
 * Function:                                                           
 *   do a full event simulation from all the pi's to po's for a good   
 *   machine with the current set of test patterns
 * Input: 
 *   pilist  - inputs with different 
 * Implicit Inputs: (variables that are global to this function)       
 *   circuit - circuit graph                                    
 *   num_pi  - number of principal inputs                       
 * Return:                                                             
 *   none
 * SideEffect:
 *   sets label0 and label1 to the new value
 */
void eventsim (CIRCUIT *circuit,
	       NODE **pilist, /* the list of pi's that have changed since last */
               	            /* simulation */
	       int  *localflist,
	       faults *flist)
{ /* eventsim */
  static int   c,m,k,i;     /* Loop counter */
  static NODE	*curgate;     /* Current gate being processed by */
                            /* event-driven simulation */
  static int   curgatefcn;  /* Function of current gate */
  static int   eval0;       /* Value of zero word for gate being */
			    /* simulated */
  static int   eval1;       /* Value of one word for gate being */
			    /* simulated */
  static int *fanoutfaults;/* a list of fanout faults for a node */
  static int *stemfaults;  /* a list of stem faults for a node */
  static int   init;        /* flag indicating whether eventsim has been */
			    /* initialized */
  int bit;
  int line;
  int ffrom;
  NODE * some_gate;

  if (init!=1)
    {
      fanoutfaults = (int *)make_ilist(MAXAFLSTLEN);
      stemfaults   = (int *)make_ilist(MAXAFLSTLEN);
      totalsfsimmemory += 2*(MAXAFLSTLEN+1)*sizeof(int);
      fanoutfaults[0] = 0;
      stemfaults[0] = 0;
      init = 1;
    }

  /* Two loops are necessary because we must enqueue everything in level */
  /* number order */
  for (k = 1; k <=(int)pilist[0]; k++) 
  //loop_through_ilist(pilist, k, m)
    {
      //some_gate = PIs (circuit, k);
      for (m = 0; m < nmbroutputs (pilist[k]); m++)
	event_enqueue (fanoutgate (pilist[k], m));
    }

/* X. Chen and Arjav Parikh made the following two lines on June 23, 1993 */
#ifdef DEBUG
   for(m=1;m<=faultcnt;m++)
      event_enqueue (get_node_pntr(circuit, m));
#endif
  while (! empty_queue())
    {
      /*
       * Propagate the gate value forwards after simulation (go BACKWARDS
       * only to get the gate inputs)
       */
      event_dequeue(curgate);
      curgatefcn = gatefunction(curgate);
      eval0 = get_label0(curgate);
      eval1 = get_label1(curgate);
      
/* If the gate has fanins */
/* don't use < pi, this is bad for symbolic processing...PeteChen 1/10/92 */
      if ((curgatefcn != p_pi) && (nmbrinputs(curgate) > 0))
	{
	  /* prior to implication, init all input lines */
	  if(gatefunction(curgate) == p_d_ff)
	    {
	      for(m=1;m<=nmbrinputs(curgate);m++)
		{
		  some_gate = faningate(curgate, m-1);
		  get_line0(get_ppo_pntr(curgate),m-1) = get_label0(some_gate);
		  get_line1(get_ppo_pntr(curgate),m-1) = get_label1(some_gate);
		  implicate_forw(get_ppo_pntr(curgate),gatefunction(get_ppo_pntr(curgate)));
		}
	    }
	  else
	    {
	      for(m=1;m<=nmbrinputs(curgate);m++)
		{
		  some_gate = faningate(curgate, m-1);
		  get_line0(curgate,m-1) = get_label0(some_gate);
		  get_line1(curgate,m-1) = get_label1(some_gate);
		}
	    }

	  if ( (get_activeflist(curgate)!=NULL) && 
	      (nmbraflist(curgate)!=0) )
	    {
	      /* 
	       * This is a node with some faults, let's process the
	       * active fault list, and separate the stem faults from
	       * the fanout faults
	       */
	      ilist_length(fanoutfaults) = 0;
	      ilist_length(stemfaults)   = 0;

	      if(gatefunction(curgate)!=p_d_ff)
		{
		  for(k=1; k<=nmbraflist(curgate);k++)
		    {
		      m=get_activeflist(curgate)[k];
		      if (activefault_to(m,localflist,flist)==0)
			add_to_itail(fanoutfaults,m);
		      else
			add_to_itail(stemfaults,m);
		    }
		  
		  /* 
		   * Inject stem faults:
		   * set the input lines to the stuck-at values of
		   * corresponding stem faults, before implication
		   */

		  for(k=1;k<=stemfaults[0];k++)
		    { /* inject stem faults */
		      bit = stemfaults[k];
		      ffrom = activefault_from(bit,localflist,flist);
		      //ffrom = (ref_tab[ffrom]!=0)? ref_tab[ffrom]: ffrom;
		      for(m=1; m<=nmbrinputs(curgate);m++)
			{
			  line = gate_number(faningate(curgate,m-1));
			  
			  if (line==ffrom)
			    break;
			}
		      
		      set_line(curgate, m, bit,activefault_sa(bit,localflist,flist));
		    }

		  implicate_forw(curgate,curgatefcn);

		  /*
		   * Inject fanout faults:
		   * set the label of the node to the stuck-at values of
		   * corresponding fanout faults after implication
		   */
		  for(k=1;k<=fanoutfaults[0];k++)
		    { /* inject fanout faults */
		      bit = fanoutfaults[k];
		      set_line(curgate,0,bit,activefault_sa(bit,localflist,flist));
		    }
		}
	      else if(gatefunction(curgate)==p_d_ff)
		{
		  for(k=1; k<=nmbraflist(curgate);k++)
		    {
		      m=get_activeflist(curgate)[k];
		      if (activefault_to(m,localflist,flist)==0)
			add_to_itail(fanoutfaults,m);
		      else
			add_to_itail(stemfaults,m);
		    }
		  
		  /* 
		   * Inject stem faults:
		   * set the input lines to the stuck-at values of
		   * corresponding stem faults, before implication
		   */

		  for(k=1;k<=stemfaults[0];k++)
		    { /* inject stem faults */
		      bit = stemfaults[k];
		      ffrom = activefault_from(bit,localflist,flist);
		      //ffrom = (ref_tab[ffrom]!=0)? ref_tab[ffrom]: ffrom;
		      for(m=1; m<=nmbrinputs(/*curgate*/get_ppo_pntr(curgate));m++)
			{
			  line = gate_number(faningate(/*curgate*/get_ppo_pntr(curgate),m-1));
			  
			  if (line==ffrom)
			    break;
			}
		      
		      set_line(/*curgate*/get_ppo_pntr(curgate), m, bit,activefault_sa(bit,localflist,flist));
		    }

		  implicate_forw(/*curgate*/get_ppo_pntr(curgate),/*curgatefcn*/gatefunction(get_ppo_pntr(curgate)));

		  /*
		   * Inject fanout faults:
		   * set the label of the node to the stuck-at values of
		   * corresponding fanout faults after implication
		   */
		  for(k=1;k<=fanoutfaults[0];k++)
		    { /* inject fanout faults */
		      bit = fanoutfaults[k];
		      set_line(/*curgate*/get_ppo_pntr(curgate),0,bit,activefault_sa(bit,localflist,flist));
		    }
		}
	    }
	  else
	    /*
	     * a node without any fault, perform straight implication
	       Don't Evaluate the FF. States are in PPOs and PPOs will
	       be copied into gs_0 and gs_1 in sfsim. Then new gs_0 &
	       gs_1 will compared to FF's labels and then its fanouts
	       will be enqueued. 
	    */
	    if(gatefunction(curgate)!=p_d_ff)
	      implicate_forw(curgate,curgatefcn);

	  if (eval0 != get_label0(curgate) || eval1 != get_label1(curgate))
	    {
	      /* Put this changed value on the event queue for every gate */
	      /* fanout, DO NOT enqueue fanouts of a DFF. Enqueue before next 
	       vector is applied.*/
	      if(gatefunction(curgate)!=p_d_ff)
		{
		  for(i=0;i<nmbroutputs(curgate);i++)
		    {
		      some_gate = fanoutgate(curgate, i);
		      event_enqueue (some_gate);
		    }
		}
	    }
	}
    } /* while */
} /* eventsim */

void print_lflist(int *lflist)
{
  int i,m,k;

  fprintf(resultsout, "Printing local faultlist(print_lflist)---\n");
  for(m=1; m<lflist[0];m++)
    {
     k = lflist[m];
     fprintf(resultsout,"Fault[%2d-%3d] %5d %5d %2d states---",
		m, k,
		faultlist[k]->from,faultlist[k]->to,
		faultlist[k]->sa);
     for(i=0;i<num_dff;i++)
       fprintf(resultsout,"%c",(faultlist[(k)]->state)[i]);
     fprintf(resultsout, "\n");
    }
}
void print_pos(int *localflist, int cur_vec )
{
  int vec, po_index, i;
  NODE	*c;
  fprintf(resultsout, "Printing the Primary Outputs(print_pos) of frame %d---", cur_vec);
  for(vec = 0; vec <= localflist[0]; vec++)
    {
      if(vec == 0) 
	{
	  fprintf(resultsout,"good mchn POs--");
	  for (i=0; i<nmbr_po(circuit); i++)
	    {
		c = POs(circuit, i);
	      po_index = ((get_label0(c)>>vec)&01) | (((get_label1(c)>>vec)<<1)&02);
	      fprintf(resultsout, "%c",trans_state[(int)po_index][0]);
	    }
	  fprintf(resultsout, "\n");
	}
      else
	{
	  fprintf(resultsout,"Fault[%2d-%4d] %s, POs--", vec, localflist[vec],
		  (tmpdetected[localflist[vec]])? "  detected":"undetected");
	  for (i=0; i<nmbr_po(circuit); i++)
	    {
		c = POs(circuit, i);
	      po_index = ((get_label0(c)>>vec)&01) | (((get_label1(c)>>vec)<<1)&02);
	      fprintf(resultsout, "%c",trans_state[(int)po_index][0]);
	    }
	  fprintf(resultsout, "\n");
	}
    }
}

/*
 * Function:
 *   Simulate a good machine and 31 faulty machines in parallel
 * Inputs:
 *   results    - a file pointer to an output file
 *   circuit	- a pointer to a circuit graph (not used)
 *   localflist - an ilist with elements indexing into flist
 *   flist      - a global fault list of all faults needed to be simulated
 *   plength    - the ordinal nubmer of sensitization frame, counted
 *                backwards as appeared in the sequence file (not used)
 *   multiflag  - a flag indicating whether this call is running in
 *                verification mode or fault dropping mode, this is used to
 *		  determine different output format
 * Implicit Inputs:
 *   num_pi  - number of pi's
 *   num_po  - number of po's
 *   num_dff - number of dff's
 * Return:
 *   number of tested faults
 * SideEffect:
 *   - sets lines, and flabel of a node for implications,
 *   - sets fault_detected for the faults detected in the flist
 *   - output simulation result to file indicated by results
 */
int sfsim (char vecfile_name[], /*name of the input vector file*/
	   int   *vec_array, /*array of vectors to be simulated */
	   int    vec_len,   /* number of bits of one vector, this
				is not always num_pi - num_dff, 
				in SEST, it's num_pi */
	   int    num_vec,   /* number of vectors to be simulated*/
	   int file_or_array,/*USEFILE: using vector file as vector souce,
			       USEARRAY: using the vector array as vector source*/
	   FILE   *results,  /* file pointer to the commentary file */
	   CIRCUIT	*circuit,      /* pointer to the circuit graph */
	   faults *flist,    /* the list of faults needed to be simulated */
	   int    findex,    /* >0: index of the fault to be simulated,
				 0: indicating fault dropping mode--
				    simulate all faults not detected 
			        <0: simulate all faults detected or undected*/
	   int  debugmode,   /* operating mode: */
                             /*  FSMODE_NORM  : '0' normal mode */
                             /*  FSMODE_DEBUG : '1' DEBUG mode */
	   int  creatflist,  /*  CREAT_FLIST  : don't creat fault list for each vector
			         NCREAT_FLIST : creat tested-fault list for each vector
				 SIM_FLIST    : simulate fault list for vector flnum*/
	   int    flnum,     /*  vector number of fault list to be simulated */
	   int  updatestate, /*  UPDATE_STATE : don't update the state list of faults,
				                for tournament
				 NUPDATE_STATE: update the state list of faults   
				 XSTATE       : simulate using X state */
	   int  updatefault, /*  UPDATE_FAULT : don't update the fault->flag even if detected
			         NUPDATE_FAULT: update the fault->flag once detected*/
	   int  initialfault,
	   /* INITIAL_FAULT    - initial all faults to be undected, states to Xs'*/
	   /* UNINITIAL_FAULT  - keep the faults detected/undetected as they are*/
	   int  cmpct,   
	   /* true/false: the vec array is compacted or not */
	   int startvec,     /*0 means the first vector*/ 
	   int earlystop)    /*EARLY_STOP: stop simulation once all faults are detected
			       NEARLY_STOP: continue to simulate until the last vector*/

{ /* sfsim */
  NODE *some_gate;
  int tested=0;                  /* Set to 1 if the current fault is */
				 /* tested, 0 otherwise */ 
  static  int vec;               /* bit position of the test pattern */
  static  int c;                 /* Loop counters */
  int     po_start_f;            /* Index of starting PO gate */
  unsigned int gm0, gm1;          /* words for good machine response */
  unsigned int fm0, fm1;          /* words for faulty machine response */
  unsigned int mask_test;         /* mask of the test patterns, the a 1 */
				 /* bit indicates a useful test pattern */
  unsigned int po_index;/* index to the pi and po translation */
				 /* table */
  //int num_vec;
  //int	gatecnt = nmbr_gates(circuit);
  int cur_vec;
  int num_rpi;
  int num_rpo;
  static int *gs_0;
  static int *gs_1;
  int i, k, m;
  int from;
  int to;
  int ffrom;
  int fto;
  int sav;
  int ppo_start;//=gatecnt+1;
  //static int *pilist;
  static NODE **pilist;
  static int *localflist;/* list of faults to be simulated for this simulation */
  int faultstart;
  int faultend;
  int fi;
  int vecuse=0;
  int flistempty;

  int tmptested=0; //J. Zhang--used to early stop the simulation

  //for debug purpose
  int sim_iter=0;
  int debug_all = false;
  int print_each_vec = true;
  int D_exist = false; // for creating D_state

  int sfsimdebugall = sfsimdebug || 1;  

#define MAXDFL ((faultcnt/10 > 100) ? faultcnt/10:100)

  if(prt_VEC_GM ==1) vectorsfile = fopen(vecfile_name, "w");

  elapsetime_s(results,TIME_SFSIM,false);

  if(file_or_array == USEFILE)
    {
      vectorsfile = fopen(vecfile_name, "r");
      position = -1; // reset the position to -1 for this sim iteration -J. Zhang 
    }
  
  if (findex!=0&&debugmode==FSMODE_DEBUG) //single fault is being simulated
    {
      if (fault_to (findex) == 0)
	fprintf(results,"Fault No. %d: %d sav %d\n",
		findex,fault_from(findex),fault_sa(findex));
      else
	fprintf(results,"Fault No. %d: %d %d sav %d\n",
		findex,fault_from(findex),fault_to(findex),fault_sa(findex));
    }
  /* gatecnt is total gates (LG+FFs+POs+PIs) */
  po_start_f = gatecnt - num_po + 1;//??
  ppo_start = gatecnt-num_dff+1;//index of the 1st dff output
  num_rpi = num_pi - num_dff; //number of real primary pi's
  num_rpo = num_po - num_dff; //number of real primary po's
  
  if (equeue_allocated == false)
    {
      queuesize = QUEUESIZE;
      eventqueue = (NODE **) malloc (QUEUESIZE * sizeof (NODE *));
      totalsfsimmemory += QUEUESIZE * sizeof(NODE *);
      equeue_allocated = true;
      /* 
       * supposed to use calloc here, but mallocPack has a broken calloc
       * which won't run correctly, so don't use calloc.  PeteChen
       * 1/10/92
       */
      //num_pi = nmbr_pi(circuit);
      //pilist = (int *)make_ilist(nmbr_pi(circuit));
      pilist = (NODE **)make_ilist(num_pi);
      for(i=0; i<num_pi; i++)
	{
	  pilist[i] = (NODE*)make_ilist(1);
	}

      pilist[0] = (int) 0;
      localflist = (int *)make_ilist(MAXAFLSTLEN);
      localflist[0] = 0;
      gs_0 = (int *)calloc(num_dff, sizeof(int));//Don't use malloc, it does not set to all 0s
      gs_1 = (int *)calloc(num_dff, sizeof(int));//Don't use malloc, it does not set to all 0s
      totalsfsimmemory += (MAXAFLSTLEN + num_pi + 2 + 2*num_dff)*sizeof(int);
    }

  if(creatflist == CREAT_FLIST)//??
    {
      alldflist[0] = (int *) 0; //NULL pointer
      state_initialized = 0;
      initial_stt_vec = 0;
      if(detectedflist == NULL)
	{
	  detectedflist = (int **)malloc((num_vec+1)*sizeof(int *));
	  detectedflist[0] = (int *) num_vec;//?? 
	  dflistsize = (int *)malloc((num_vec+1)*sizeof(int));
	  for(i=1;i<=num_vec;i++)
	    {
	      detectedflist[i] = (int *)make_ilist(50);
	      ilist_length(detectedflist[i]) = 0;
	      dflistsize[i]=50;
	      // printf("df[%d] length---%d\n", i, ilist_length(detectedflist[i]) );
	    }
	  totalsfsimmemory += num_vec*( sizeof(int *) + sizeof(int)*52 );
	}
      //resize the space of detectedflist if it is not NULL
      else // (detectedflist != NULL)
	{ 
	  c = (int) detectedflist[0];
	  if(num_vec > c)
	    {
	      detectedflist = (int **)realloc(detectedflist,(num_vec+1)*sizeof(int *));
	      detectedflist[0] = (int *) num_vec; 
	      dflistsize = (int *)realloc(dflistsize, (num_vec+1)*sizeof(int));
	      if(detectedflist == NULL || dflistsize == NULL)
		{
		  fprintf(resultsout, "memory realloc wrong!!--sfsim(a)\n");
		  exit(0);
		}	      
	      totalsfsimmemory += (num_vec-c)*( sizeof(int *) + sizeof(int)*52 );
	    }
	  for(i= 1;i<=num_vec;i++)
	    {
	      if(i>c)
		{
		  detectedflist[i] = (int *)make_ilist(50);
		  dflistsize[i]=50;
		  if(detectedflist[i] == NULL)
		    {
		      fprintf(resultsout, "memory alloc wrong!!--sfsim(b)\n");
		      exit(0);
		    }
		}
	      ilist_length(detectedflist[i]) = 0;
	    }
	  // printf("df[%d] length---%d\n", i, ilist_length(detectedflist[i]) );
	}      
    }
  
  /*First initialize all faults according to "initialfault" */
  if(initialfault == INITIAL_FAULT)
    for (fi=1;fi <=faultcnt; fi++)
      {
	fault_detected(fi) = false;
	tmpdetected[fi] = false;
      }
  else if(initialfault == UNINITIAL_FAULT) 
    for (fi=1;fi <=faultcnt; fi++)
      {
	tmpdetected[fi] = fault_detected(fi);
      }

  /*if D_state needs to be created*/
  if(creat_D_state == true)//??what is this doing 
    {
      for (fi=1; fi<=faultcnt; fi++)
	if(fault_detected(fi) == false)
	  {
	    printf("F %5d -sfsim\n ", fi);
	    faultlist[fi]->D_state = (int**)malloc((num_vec+1)*sizeof(int*));
	    (faultlist[fi]->D_size) = 0;
	  }
    }
  
  /* Select faults to be simulated according findex */
  if(creatflist != SIM_FLIST)
    {
      if(findex > 0) faultstart = faultend = findex;
      else if(findex < 0) 
	{      
	  faultstart = 1;
	  faultend   = faultcnt;
	}
      else // findex == 0: fault droping mode
	{      
	  faultstart = 1;
	  faultend   = faultcnt;
	  //make sure the fault[faultend] undetected
	  while(fault_detected(faultend) == true && faultend > 1)
	    faultend--;
	  //the following case means that no fault need to be simulated
	  if(faultend == 1 && fault_detected(1) == true)
	    faultend = 0;
	}
    }
  else // creatflist == SIM_FLIST
    {
      if( flnum < 1 ) 
	{
	  //fprintf(results, "flnum = %d, Wrong!!-sfsim\n", flnum);
	  goto finished;
	}
      else if( ilist_length(detectedflist[flnum]) < 1 ) 
	{
	  //fprintf(results, "length of detectedflist[%d]=%d, compaction was skipped!!-sfsim\n",
	  //	  flnum,  ilist_length(detectedflist[flnum]));
	  goto finished;
	}
      else
	{
	  faultstart = 1;
	  faultend   = ilist_length(detectedflist[flnum]);
	}
      if(sfsimdebug)
	if( (file_or_array != USEARRAY) || (flnum < 1) || 
	    ((updatestate != NUPDATE_STATE) && (updatestate != XSTATE )) ||
	    (updatefault != NUPDATE_FAULT) || (cmpct != true))
	  {
	    fprintf(results, "Suggested Useage of SIM_FLIST: \n
                 file_or_array = USEARRAY ;      \n
                         flnum > 0        ;      \n
	           updatestate = NUPDATE_STATE or Xstate; \n
	           updatefault = NUPDATE_FAULT ; \n
                         cmpct = true;           \n");
	  }
    }
  
  /* loop through all faults to be simulated */
  while (faultstart<=faultend)
    {
      sim_iter ++;
      if(sfsimdebugall) fprintf(results, "sfsim[1]:sim_iter=%d-faultstart=%d-faultend=%d\n"
			     ,sim_iter,faultstart,faultend);
      ilist_length(localflist) = 0;
      for (fi=faultstart;fi <=faultend; fi++)
	{
	  /*findex < 0 means simulate faults regardless of being detected/undetected
	    -J. Zhang  */
	  if(creatflist != SIM_FLIST)
	    {
	      if (fault_detected(fi)!=true || findex < 0) 
		add_to_itail(localflist,fi);
	    }
	  else // creatflist == SIM_FLIST
	    {
	      add_to_itail(localflist,detectedflist[flnum][fi]);
	      tmpdetected[(detectedflist[flnum][fi])] = false;
	    }
	  if (localflist[0]>=MAXAFLSTLEN)
	    break;
	}
      tmptested = 0;
      faultstart = fi+1;
      if(sfsimdebugall) 
	{
	  fprintf(results, "sfsim[2]:sim_iter=%d-local faults:\n",sim_iter);
	  for(i=1; i<=localflist[0];i++)  
	     fprintf(results, " %d", localflist[i]);
	  fprintf(results, "\n");
	}

      /* the list is empty, some logic error occurs, quit*/
      if (localflist[0]==0)
	{
	  flistempty = true; //so far not used
	  fprintf(results,"length of localflist = 0!! wrong!! \n");
	  printf("length of localflist = 0!! wrong!! \n");
	  if(file_or_array == USEFILE) fclose(vectorsfile);
	  return (-1);
	}      
      
      /* attach faults to corresponding gates */
      //for(i=1;i<=localflist[0];i++)/*i<localflist[0] to i<=localflist[0] -- ok */
      loop_through_ilist(localflist,i,k)
	{
	  //k = localflist[i];
	  from  = lfault_from(flist,k);
	  to    = lfault_to(flist,k);
	  sav   = lfault_sa(flist,k);
	  /* WARNING -- Must enable to call to 'readfaults' */
	  /* WARNING -- Need to construct ref_tab, it is in 'cktconvert.c' */
	  //ffrom = (ref_tab[from]!=0)? ref_tab[from]: from;
	  //fto = cfaultto(to);
	  
	  /* for fanout faults, attach fault to from gate */
	  if (/*fto*/to==0) /* If fto=0, means fault is a stem fault */ 
	    {
	      some_gate = get_node_pntr(circuit, /*ffrom*/from);
	      if(some_gate->aflist == NULL)
		{
		  some_gate->aflist = (int *)make_ilist(MAXAFLSTLEN);
		  totalsfsimmemory += (MAXAFLSTLEN+1)*sizeof(int);
		  some_gate->nmbr_aflist = 0;
		  some_gate->aflist[0] = 0;
		}
	      add_to_itail(some_gate->aflist,i);
	    }
	  // non-fanout faults, attach fault to the to gate --J.ZHANG
	  else
	    {
	      some_gate = get_node_pntr(circuit, /*fto*/to);
	      if(some_gate->aflist == NULL)
		{
		  some_gate->aflist = (int *) make_ilist(MAXAFLSTLEN);
		  totalsfsimmemory += (MAXAFLSTLEN+1)*sizeof(int);
		  some_gate->nmbr_aflist = 0;
		  some_gate->aflist[0] = 0;
		}
	      add_to_itail(some_gate->aflist,i);
	    }
	}
      
      /* clear all previous gatelabeling--set all gate "X" */
      clearLabels(circuit);

      if(sfsimdebugall)
	{
	  fprintf(results, "sfsim[3]:sim_iter=%d-initialfault=%d-local faults' states before state init--:\n"
		  ,sim_iter,initialfault);
	  print_lflist(localflist);
	}
      /* initialize all dff to correct state according to 'initialfault'*/ 
      if(sfsimdebugall)
	{
	  fprintf(results, "sfsim[4]:sim_iter=%d--DFF words(gs_0&gs_1) BeFore & AFter state injection-------:\n" ,sim_iter);
	}

      /* WARNING -- nmbrppos(circuit) is not set to num_dff, must fix in r_parse */
      //for (i=0;i<nmbrppos(circuit);i++)
      for (i=0;i<num_dff;i++)
	{
	  if(sfsimdebugall)
	    {
	      fprintf(results, "BeFore--DFF[%d]:", i);
	      for(k=0;k<32;k++) 	 
		{      
		  po_index = ((gs_0[i]>>k)&01) | (((gs_1[i]>>k)<<1)&02);
		  fprintf(results,"%c",trans_state[(int)po_index][0]);
		  if(k%5 == 4) fprintf(results, " ");
		  if(k == 31) fprintf(results, "\n");
		}
	    }	  
	  if(initialfault == INITIAL_FAULT)//initial good state = 'X'
	    (goodstate)[i] = 'X';//??

	  c = goodstate[i];
	  if(updatestate == XSTATE) c = 'X';

	  switch(c)//??
	    {
	    case 'X':
	      if(sfsimdebugall && debug_all)
		fprintf(results, "good state=%c--case X\n",(goodstate)[i]);
	      clear_bit (gs_0[i],0);
	      clear_bit (gs_1[i],0);
	      break;
	    case '0':
	      if(sfsimdebugall && debug_all)
		fprintf(results, "good state=%c--case 0\n",(goodstate)[i]);
	      set_bit   (gs_0[i],0);
	      clear_bit (gs_1[i],0);
	      break;
	    case '1':
	      if(sfsimdebugall && debug_all)
		fprintf(results, "good state=%c--case 1\n",(goodstate)[i]);
	      clear_bit (gs_0[i],0);
	      set_bit   (gs_1[i],0);
	      break;
	    default:
	      clear_bit (gs_0[i],0);
	      clear_bit (gs_1[i],0);
	    }
	  //for(m=1; m<localflist[0];m++)
	  loop_through_ilist(localflist,m,k)//?? why are we looping again
	    {	  
	      //k = localflist[m];
	      if(initialfault == INITIAL_FAULT)//initial faulty state = 'X'
		((flist)[(k)]->state)[i] = 'X';
	      
	      if(sfsimdebugall && debug_all) 
		fprintf(resultsout, "f[%2d-%3d][%2d]=%c--",m,k,i,((flist)[(k)]->state)[i]);

	      c = ((flist)[(k)]->state)[i];//??
	      if(updatestate == XSTATE) c = 'X';
	      
	      switch(c)
		{
		case 'X':
		  if(sfsimdebugall && debug_all)
		    fprintf(resultsout, "faulty state--case X \n");
		  clear_bit (gs_0[i],m);
		  clear_bit (gs_1[i],m);
		  break;
		case '0':
		  if(sfsimdebugall && debug_all)
		    fprintf(resultsout, "faulty state--case 0 \n");
		  set_bit   (gs_0[i],m);
		  clear_bit (gs_1[i],m);
		  break;
		case '1':
		  if(sfsimdebugall && debug_all)
		    fprintf(resultsout, "faulty state--case 1 \n");
		  clear_bit (gs_0[i],m);
		  set_bit   (gs_1[i],m);
		  break;
		default:
		  if(sfsimdebugall && debug_all)
		    fprintf(resultsout, "faulty state--case DEFAULT \n");
		  exit(-1);
		  clear_bit (gs_0[i],m);
		  clear_bit (gs_1[i],m);
		}
	    }
	  if(sfsimdebugall)
	    {
	      fprintf(results, "AFter---DFF[%d]:", i);
	      for(k=0;k<32;k++) 	 
		{      
		  po_index = ((gs_0[i]>>k)&01) | (((gs_1[i]>>k)<<1)&02);
		  fprintf(results,"%c",trans_state[(int)po_index][0]);
		  if(k%5 == 4) fprintf(results, " ");
		  if(k == 31) fprintf(results, "\n");
		}
	    }
	}//for(i=0;i<num_dff;..
      
      if(sfsimdebugall){
	fprintf(results, "sfsim[5]:sim_iter=%d-initialfault=%d-local faults' states after initialization--:\n"
		,sim_iter,initialfault);
	print_lflist(localflist);
      }
      
      /*
       * loop through all the vectors in the sequence backwards, 
       * because sest is using reverse time processing 
       * ---I don't understand the "reverse time processing" part -J.Zhang
       */
  vecuse = startvec-1;
  // if(num_vec >= 1);
  for (cur_vec = 1; cur_vec <= num_vec; cur_vec = cur_vec + 1)
    {
      //using compacted vectors only, this is effective just for array now -J.Zhang
      if(cmpct == true && file_or_array == USEARRAY ) 
	{
	  vecuse++;
	  if( removed == NULL ) 
	    {
	      fprintf(results, "removed is NULL!!!-Wrong!!!-sfsim\n");
	      printf("removed is NULL!!!-Wrong!!!-sfsim\n");
	      return (-1);	      
	    }	 
	  else 
	    while(removed[vecuse] == 1) 
	      {
		/*if(sfsimdebugall) 
		  fprintf(results, "vecuse = %d removed[vecuse]=%d !!!!!!\n", 
		  vecuse, removed[vecuse]);*/
		vecuse++;
	      }
	}
      //using the vec file or uncompacted vector array
      else vecuse = cur_vec-1;
      
      pilist[0] = (int) 0;
      
      //for(i=1;i<localflist[0];i++)
      loop_through_ilist(localflist,i,k)//??why are we doing this again
	{
	  //k = localflist[i];
	  from  = lfault_from(flist,k);
	  to    = lfault_to(flist,k);
	  sav   = lfault_sa(flist,k);
	  //ffrom = (ref_tab[from]!=0)? ref_tab[from]: from;
	  //fto   = cfaultto(to);
	  
	  /* if fault is in the primary input inject it */
	  if (/*fto*/to==0 && gatefunction(get_node_pntr(circuit,/*ffrom*/from))==p_pi) 
	    set_line(get_node_pntr(circuit,/*ffrom*/from),0,i,sav);
	}
      
      /* inject the test vector */
      //for (i=0;i<nmbr_pi(circuit);i++)
      for (i=1;i<=num_rpi;i++)
	{
	  some_gate = PIs(circuit, i-1);
	  if(sfsimdebugall) 
	    if(i==0) fprintf(results, "sfsim[7]:vec[%d]---", vecuse+1);  
	  
	  /*Here we select the proper input vector according to "file_or_array*/
	  if(file_or_array == USEARRAY) 
	    {
	      //fprintf(resultsout, "array");
	      c = vec_array[vecuse*vec_len+i];
	    }
	  else  /* Read from Vectors File */
	    {
	      if(vectorsfile == NULL) fprintf(resultsout, "vectorsfile is NULL!!!-Wrong!!-sfsim\n");
	      //fprintf(results, "position = %d, ", position);
	      c = get_testseq_bit(num_rpi);
	    }
	  if(sfsimdebugall)
	    {
	      fprintf(results, "%c",c);
	      //if(i==nmbr_pi(circuit)-1) fprintf(results, "\n");
	      if(i==num_pi) fprintf(results, "\n");
	    }
	  if(prt_VEC_GM==1) fprintf(vectorsfile, "%c",c );

	  switch(c)
	    {
	    case '1': //ONE:
	      gm0 = 0;
	      gm1 = ~0;
	      /* fprintf (results, "Inject vector pi index %d = 1\n", i);  MB */
	      break;
	    case '0': //ZERO:
	      gm0 = ~0;
	      gm1 = 0;
	      /*fprintf (results, "Inject vector pi index %d = 0\n", i);  MB */
	      break;
	    case 'X': //X:
	      gm0 = 0;
	      gm1 = 0;
	      /*fprintf (results, "Inject vector pi index %d = X\n", i);  MB */
	      break;
	    default:
	      fprintf (results, "sfsim--Wrong!! Inject vector[%d], pi[%d] = unknown->%d\n", vecuse+1,i,c); 
	      gm0 = 0;
	      gm1 = 0;
	      exit(0);
	    }
	  
	  if (get_label0(some_gate)!=gm0||get_label1(some_gate)!=gm1)
	    {
	      get_label0(some_gate)=gm0;
	      get_label1(some_gate)=gm1;
	      //add_to_itail(pilist,i+1);
	      add_to_itail(pilist,/*i*/PIs(circuit, i-1));
	    }
	}//for(i=1; i<=num_rpi; i++)
      
      /* inject good state vector */
      /*
	Actually, this injection is for the good machine and all faulty
	machines.
	Comments by X. Chen, Jan. 19, 1994.
      */
      /***** take DFF input states and put it in label ******/
      if(sfsimdebugall && print_each_vec )
	{
	  fprintf(results, "sfsim[8]:sim_iter=%d-vec[%d]-DFF (label0&label1) before updating--\n"
		  ,sim_iter,vecuse+1);
	  print_state(0,-1,-1, localflist, 1);  
	}      
      //for (i=0;i<num_dff;i++)
      k = 0;
      for (i=0;i<nmbr_gates(circuit);i++)
	{
	  //k = num_rpi+i+1;
	  if(gatefunction(/*gatesx*/gatesx_dup(circuit, i)) == p_d_ff)
	     {
	       if(k>=num_dff)
		 {
		   printf("ERROR -- assigning DFF input states ans putting it in label\n");
		   exit(-1);
		 }
	       if (get_label0(/*gatesx*/gatesx_dup(circuit,i))!=gs_0[k]||get_label1(/*gatesx*/gatesx_dup(circuit, i))!=gs_1[k])
	  //if (get_label0(gatesx(circuit,k-1))!=gs_0[i]||get_label1(gatesx(circuit, k-1))!=gs_1[i])
		 {
		   get_label0(/*gatesx*/gatesx_dup(circuit, i)) = gs_0[k];
		   get_label1(/*gatesx*/gatesx_dup(circuit, i)) = gs_1[k];
		   add_to_itail(pilist,/*gatesx*/gatesx_dup(circuit, i));
		 }
	       k++;
	     }
	}
      k= 0; // reset k just in case it is being used below
      if(sfsimdebugall && print_each_vec)
	{
	  fprintf(results, "sfsim[9]:sim_iter=%d-cur_vec=%d-DFF (label0&label1) after updating--\n"
		  ,sim_iter,vecuse+1);
	  print_state(0,-1,-1, localflist, 1);  
	}      
      
      //inject faults from pi
      //for(i=1;i<=localflist[0];i++)
      loop_through_ilist(localflist,i,k)//?? why loop again?
	{
	  //k = localflist[i];
	  from  = lfault_from(flist,k);
	  to    = lfault_to(flist,k);
	  sav   = lfault_sa(flist,k);
	  //ffrom = (ref_tab[from]!=0)? ref_tab[from]: from;
	  //fto   = cfaultto(to);
	  
	  if (/*fto*/to==0 && gatefunction(get_node_pntr(circuit, /*ffrom*/from))==p_pi)//??pi->p_pi 
	    set_line(get_node_pntr(circuit, /*ffrom*/from),0,i,sav);
	}
      
      eventsim(circuit,pilist,localflist,flist);
      
      /* print the good machine responses (both POs and DFFs) */
      if(prt_VEC_GM==1) 
	{

	  fprintf(vectorsfile, "   " );
	  for ( c=po_start_f; c<ppo_start; c++) //possible bug, po_start_f is 21, loop not entered.
	    {
	      po_index = ((get_label0(get_node_pntr(circuit, c)))&01) | ((get_label1(get_node_pntr(circuit, c))<<1)&02);
	      fprintf(vectorsfile, "%c", trans_state[(int)po_index][0]);
	    }
	  fprintf(vectorsfile, "\n" );
	}

      /* save the state vector for the good machine */
      /*
	Actually, this is to save the state vector for the good machine
	and all faulty machines.  
	Comments by X. Chen, Jan. 19, 1994.
      */
      
      /******** put back the label in good state from ppos ********/
      for ( c=ppo_start; c<=gatecnt; c++)//Possible bug!. PPO Labels were never set after sim.
	{
	  gs_0[c-ppo_start] = get_label0(PPOs(circuit, c-ppo_start)/*get_node_pntr(circuit, c)*/);
	  gs_1[c-ppo_start] = get_label1(PPOs(circuit, c-ppo_start)/*get_node_pntr(circuit, c)*/);
	  /*	      if (((label0 (c) & 01) != 0) &&
		      ((label1 (c) & 01) == 0))
		      fprintf (results, "Next state line flip-flop index %d = 0\n", c);  MB 
		      else if (((label0 (c) & 01) == 0) &&
		      ((label1 (c) & 01) != 0))
		      fprintf (results, "Next state line flip-flop index %d = 1\n", c);  MB 
		      else if (((label0 (c) & 01) == 0) &&
		      ((label1 (c) & 01) == 0))
		      fprintf (results, "Next state line flip-flop index %d = X\n", c);  MB */

	  /*check-point the initial state vector sequence */
	  if( (creatflist == CREAT_FLIST) && (initial_stt_vec == 0) && (faultstart>faultend) )
	    {
	      po_index = ((get_label0(PPOs(circuit, c-ppo_start)/*get_node_pntr(circuit, c)*/))&01) | ((get_label1(PPOs(circuit, c-ppo_start)/*get_node_pntr(circuit, c)*/)<<1)&02);
	       if( trans_state[(int)po_index][0] != 'X') state_initialized ++;
	       if( (c==gatecnt) )
		 {
		   if( state_initialized==num_dff )
		     initial_stt_vec = vecuse + 1;
		   state_initialized = 0;//reset it before simulating the next vector
		 } 
	    }	  
	  /* Save the states according to "updatestate" -J.ZHANG*/
	  if((cur_vec == num_vec) && (updatestate == UPDATE_STATE) )
	    {
	      //update states for each fault being simulated 
	      for(i=1;i<=localflist[0];i++)
		{
		  k = localflist[i];
		  po_index = (((get_label0(PPOs(circuit, c-ppo_start)/*get_node_pntr(circuit, c)*/))>>i)&01) | (((get_label1(PPOs(circuit, c-ppo_start)/*get_node_pntr(circuit, c)*/)>>i)<<1)&02);
		  (flist[(k)]->state)[c-ppo_start] = trans_state[(int)po_index][0]; 
		}
	      //if all faults are simulated, update goodstate
	      if( faultstart > faultend ) 
		{
		  po_index = ((get_label0(PPOs(circuit, c-ppo_start)/*get_node_pntr(circuit, c)*/))&01) | ((get_label1(PPOs(circuit, c-ppo_start)/*get_node_pntr(circuit, c)*/)<<1)&02);
		  (goodstate)[c-ppo_start] = trans_state[(int)po_index][0]; 
		}
	    }//if(cur_vec...
	}//for(c=ppo_start...

      /*creat D state*/
      if(creat_D_state == true)
	loop_through_ilist(localflist,i,k)
	{
	  D_exist = false;
	  for ( c=ppo_start; c<=gatecnt; c++)
	    {	      
	      po_index = (get_label0(get_node_pntr(circuit, c))&01) |
		((get_label1(get_node_pntr(circuit, c)))<<1&02) |
		((get_label0(get_node_pntr(circuit, c))>>i)<<2&04) |
		((get_label1(get_node_pntr(circuit, c))>>i)<<3&010);
	      (flist[(k)]->tmp_state)[c-ppo_start] = po_index 
		= trans_po [(int)po_index][0];  
	      if( po_index == 'd' || po_index == 'D')
		{
		  D_exist = true;
		} 
	    }
	  
	  if ( D_exist == true ) 
	    {
	      po_index = ( flist[(k)]->D_size ) +1;
	      printf("cur_vec = %d , k = %d, po_index = %d \n", cur_vec, k, po_index);

	      flist[(k)]->D_size = po_index;
	      if( (po_index % 150) == 1)
		for(m = po_index; m< po_index + 150; m++)
		  {
		  (flist[(k)]->D_state)[m] = (int*)malloc((num_dff+2)*sizeof(int));
		  if( (flist[(k)]->D_state)[m] == NULL ) printf("no enough memory!!-sfsim\n");
		  }
	      (flist[(k)]->D_state)[po_index][0] = cur_vec;
	      if (cur_vec == 0) printf("cur_vec==0!!! sfsim...\n");
	      for ( c=ppo_start; c<=gatecnt; c++)
		{	      
		  (flist[(k)]->D_state)[po_index][c-ppo_start+1] = 
		    (flist[(k)]->tmp_state)[c-ppo_start];
		}
	    }
	  // else printf("No D/Dbar found for F[%d] at vec %d\n", k, cur_vec);

	}
      
      if(sfsimdebugall && print_each_vec)
	if(cur_vec == num_vec) 
	  {      
	    fprintf(results, "sfsim[10]:sim_iter=%d-Printing local fault states after vec[%d]----\n"
		    ,sim_iter, vecuse+1);
	    print_lflist(localflist);
	  }
      
      //print good pos and faulty pos -J. Zhang
      if(sfsimdebugall){	 
	fprintf(results, "sfsim[11]:sim_iter=%d--Printing POs after sim vec[%d]----\n"
		,sim_iter,vecuse+1);
	print_pos(localflist,cur_vec);
      }//if(sfsimdebugall)
      
      /* see if this single fault is tested */
      
      /* scan po's, looking for the presence of d or dbar */
      for (c = (po_start_f); (c < po_start_f + num_rpo); c++)
	{
	  gm0 = (get_label0(get_node_pntr(circuit, c))&01)? ~0 : 0;
	  gm1 = (get_label1(get_node_pntr(circuit, c))&01)? ~0 : 0; 
	  fm0 = get_label0(get_node_pntr(circuit, c));
	  fm1 = get_label1(get_node_pntr(circuit, c));
	  mask_test = (gm0^fm0 ) & (gm1^fm1) & (fm0^fm1);
	  
	  if (mask_test != 0)
	    {
	      /* 
	       * figure out which pattern catches the fault.  note that
	       * vec will have the position of the test pattern that
	       * catches the fault.  Since the for loop is broken, once the
	       * position if found, the increment at the end of the for loop
	       * did not occur. 
	       */
	      for (vec = 1; (vec <=ilist_length(localflist)); vec++)
		{
		  //fault[localflist[vec]] is detected
		  if ((mask_test >> vec) & 01 )
		    {
		      if (vec>WORDSIZE-1) 
			{
			  fprintf(stderr,
				  "sfsim: set bit found at bit%d, panic!\n",
				  vec);
			  exit (-1);
			}
		      //A new fault is detected
		      if( tmpdetected[(localflist)[vec]] != true )
			{
			  tmpdetected[(localflist)[vec]] = true;
			  tested++;
			  tmptested ++;
			  if(earlystop == EARLY_STOP && updatestate != UPDATE_STATE
			     && tmptested == ilist_length(localflist))
			    cur_vec = num_vec;//to early stop the simulation
			  if(updatefault == UPDATE_FAULT )
			    {
			      activefault_detected(vec,localflist,flist) = true;
			    }
			  if(creatflist == CREAT_FLIST)
			    {
			      if(ilist_length(detectedflist[cur_vec]) == dflistsize[cur_vec])
				{
				  dflistsize[cur_vec] += 50;
				  detectedflist[cur_vec] = 
				    remake_ilist(detectedflist[cur_vec],dflistsize[cur_vec]);
				  totalsfsimmemory += 50;
				  if(detectedflist[cur_vec] == NULL)
				    {
				      fprintf(resultsout, 
					      "memory realloc wrong!!--sfsim(c)\n");
				      exit(0);
				    }

				  /* printf("detected flist[%d] length=%d, newsize=%d!!-sfsim\n",
			       	 cur_vec,detectedflist[cur_vec][0],dflistsize[cur_vec]);
			      fprintf(results,"detected flist[%d] length=%d, newsize=%d!!-sfsim\n",
			      cur_vec,detectedflist[cur_vec][0],dflistsize[cur_vec]);*/
			      
    				}
			      add_to_itail(detectedflist[cur_vec],(localflist)[vec]);
			      add_to_itail(alldflist,(localflist)[vec]);
			      // printf("detected flist[%d] length = %d\n",cur_vec,detectedflist[cur_vec][0]);
			      
			      if(sfsimdebugall && print_each_vec){
				fprintf(results, "sfsim[20]:add fault[%d] to detectedflist[%d]\n",
					localflist[vec],cur_vec);
				if(c == (po_start_f + num_rpo -1))
				  loop_through_ilist(detectedflist[cur_vec], i, k)
				    fprintf(results, "%d\n",k);
			      }
			      
			      // remake_ilist(detectedflist[cur_vec],ilist_length(detectedflist[cur_vec])+1);
			      // totalsfsimmemory += 1;
			    } 
			  //print the fault, po values when it is dectected for the first time 
			  if (1/*sfsimdebug == TRUE*/ /*&& print_each_vec*/)
			    {
			      fprintf(results,
				      /*"Fault[%2d-%4d] %5d %5d %2d in frame %2d",
					vec, localflist[vec],*/
				      "[%5d %5d %2d] in frame %2d",
				      activefault_from(vec,localflist,flist),
				      activefault_to(vec,localflist,flist),
				      activefault_sa(vec,localflist,flist),
				      cur_vec); 
			      
			      /* Write the output circuit response onto the file */
			      /* use the 4-bit code to index into the translation table */
			      
			      //print the po values for the newly detected fault
			      //if( tournamentmode != TOURNAMENT )
			      if(sfsimdebug && print_each_vec)
				{
				  fprintf(results, "---");
				  for (m = po_start_f; m < ( po_start_f + num_rpo); m++)
				    {
				      po_index = (get_label0(get_node_pntr(circuit, m))&01) |
					((get_label1(get_node_pntr(circuit, m)))<<1&02) |
					((get_label0(get_node_pntr(circuit, m))>>vec)<<2&04) |
					((get_label1(get_node_pntr(circuit, m))>>vec)<<3&010);
				      fprintf(results, "%s", 
					      trans_po [(int)po_index]);  
				    }//for(m=...
				}//if(sfsimdebug)
			      fprintf(results, "\n");
			    }//if(VERBOSE
			  
			}//if(tmpdetected[(localflist)...
		    }//if((mask_test>>vec...
		}//for(vec=1..
	    } //if(mask_test!=0..
	}//for(c=(po_start_f..
      
      //print out states in the last frame for the undetected faults--J.Zhang
      if( cur_vec==num_vec )
	{
	  for (vec = 1; (vec <=ilist_length(localflist)); vec++)
	    {	  
	      if( tmpdetected[localflist[vec]] != true )
		{				
		  if(sfsimdebugall)
		    {
		      fprintf(results,
			      "Fault %5d %5d %2d not detected till frame %2d---States: ",
			      activefault_from(vec,localflist,flist),
			      activefault_to(vec,localflist,flist),
			      activefault_sa(vec,localflist,flist),
			      cur_vec); 
		    }
		  for(m=ppo_start; m<=gatecnt; m++)
		    {
		      po_index = ((get_label0(PPOs(circuit, m-ppo_start))>>vec)&01) |
			(((get_label1(PPOs(circuit, m-ppo_start))>>vec)<<1)&02);
		      /* if( updatestate == UPDATE_STATE )
			(flist[localflist[vec]]->state)[m-ppo_start] 
			= trans_state[(int)po_index][0];*/ 		      
		      if(sfsimdebugall) 
			{
			  fprintf(results, "%c",trans_state[(int)po_index][0]);
			  if(m==gatecnt) fprintf(results, "\n");
			}
		    }
		}//if(tmpdetected[localflist[vec]]
	    }//for(vec = 1;...
	}//if(cur_vec == num_vec)			  	      
    }//for (cur_vec = 1; cur_vec <= num_vec; cur_vec = cur_vec + 1)...

  if(prt_VEC_GM == 1) 
    {
      fclose(vectorsfile);
      goto finished;
    }
  //clear all the current faults being simulated  
  for(i=1;i<localflist[0]; i++)
    {
      k = localflist[i];
      from  = lfault_from(flist,k);
      to    = lfault_to(flist,k);
      ffrom = (ref_tab[from]!=0)? ref_tab[from]: from;
      fto   = cfaultto(to);
      
      /* for fanout faults, attach fault to from gate */
      if (/*fto*/to==0) 
	{
	  some_gate = get_node_pntr(circuit, /*ffrom*/from);
	  //if (some_gate->activeflist!=NULL)//line 1353
	  if (get_activeflist(some_gate)!=NULL)
	    //ilist_length(activeflist(ffrom))=0;//line 1354
	    ilist_length(get_activeflist(some_gate))=0;
	}
      else
	{
	  some_gate = get_node_pntr(circuit, /*fto*/to);
	  //if (some_gate->activeflist!=NULL)
	  if (get_activeflist(some_gate)!=NULL)
	    //ilist_length(some_gate->activeflist)=0;
	     ilist_length(get_activeflist(some_gate))=0;
	}
    }//loop_through_ilist(localflist, i, k) 
  
  if(file_or_array == USEFILE ) reset_test_seq (vecfile_name);
  
}//while(faultstart<=faultend)
 finished:
    if(file_or_array == USEFILE) fclose(vectorsfile);
    // if(cmpct == false)  print_fdetected(updatefault);

    elapsetime_e(results,TIME_SFSIM,false);
    
    return (tested);
} /* sfsim */


